#include <stdio.h>
#include <stdlib.h>

#include "simlib.h"
#include "sim4.h"
#include "sim4b1.h"
#include "Xtend1.h"
#include "scriptlib.h"

#ifndef __lint
/*@unused@*/
static const char rcsid[] =
"$Id: scriptlib.c,v 1.8 1998/05/15 19:12:50 florea Exp florea $";
#endif

static int snake(int k, int x, int endx, int endy);
static int rsnake(int k, int x, int startx, int starty, int M);


void     path(int i1, int j1, int type1, int i2, int j2, int type2, 
              int dist, edit_script **head, edit_script **tail)
{
        int *SS, *DD, *II;       /* Forward vectors */
        int *RS, *RD, *RI;       /* Backward vectors */

        edit_script *head1, *tail1, *head2, *tail2;
        int midc, rmidc;
        int start, lower, upper;
        int rstart, rlower, rupper;
        int c, k, t1, t2, t;
        int maxint;
        int mi, mj, mtype;
        char flag;

        /* Boundary cases */
        if (i1 == i2) {
           if (j1 == j2) *head = NULL;
           else {
                head1 = (edit_script *) ckalloc(sizeof(edit_script));
                head1->op_type = INSERT;
                head1->num = j2-j1;
                head1->next = NULL;
                *head = *tail = head1;
           }
           return;
        }

        if (j1 == j2) {
                head1 = (edit_script *) ckalloc(sizeof(edit_script));
                head1->op_type = DELETE;
                head1->num = i2-i1;
                head1->next = NULL;
                *head = *tail = head1;
                return;
        }

        if (dist <= 1) {
           if (j2-i2 == j1-i1) {
                head1 = (edit_script *) ckalloc(sizeof(edit_script));
                head1->op_type = SUBSTITUTE;
                head1->num = i2-i1;
                head1->next = NULL;
                *head = *tail = head1;
           } else if (j2-i2 > j1-i1) {
                if (type1 == INSERT) {
                        head1 = (edit_script *) ckalloc(sizeof(edit_script));
                        head1->op_type = INSERT;
                        head1->num = 1;
                        head2 = (edit_script *) ckalloc(sizeof(edit_script));
                        head2->op_type = SUBSTITUTE;
                        head2->num = i2-i1;
                } else {
                        head1 = (edit_script *) ckalloc(sizeof(edit_script));
                        head1->op_type = SUBSTITUTE;
                        head1->num = i2-i1;
                        head2 = (edit_script *) ckalloc(sizeof(edit_script));
                        head2->op_type = INSERT;
                        head2->num = 1;
                }
                head1->next = head2;
                head2->next = NULL;
                *head = head1;
                *tail = head2;
           } else if (j2-i2 < j1-i1) {
                if (type1 == DELETE) {
                        head1 = (edit_script *) ckalloc(sizeof(edit_script));
                        head1->op_type = DELETE;
                        head1->num = 1;
                        head2 = (edit_script *) ckalloc(sizeof(edit_script));
                        head2->op_type = SUBSTITUTE;
                        head2->num = j2-j1;
                } else {
                        head1 = (edit_script *) ckalloc(sizeof(edit_script));
                        head1->op_type = SUBSTITUTE;
                        head1->num = j2-j1;
                        head2 = (edit_script *) ckalloc(sizeof(edit_script));
                        head2->op_type = DELETE;
                        head2->num = 1;
                }
                head1->next = head2;
                head2->next = NULL;
                *head = head1;
                *tail = head2;
           }
           return;
        }

        /* Divide the problem at the middle cost */
        midc = dist/2;
        rmidc = dist - midc;

        /* Compute the boundary diagonals */
        start = j1 - i1;
        lower = max(j1-i2, start-midc);
        upper = min(j2-i1, start+midc);
        rstart = j2-i2;
        rlower = max(j1-i2, rstart-rmidc);
        rupper = min(j2-i1, rstart+rmidc);

        /* Allocate space for forward vectors */
        SS = (int *)ckalloc((upper-lower+1)*sizeof(int)) - lower;
        DD = (int *)ckalloc((upper-lower+2)*sizeof(int)) - lower;
        II = (int *)ckalloc((upper-lower+2)*sizeof(int)) - lower + 1;

        /* Forward computation */
        for (k=lower; k<=upper; ++k) SS[k] = MININT;
        for (k=lower; k<=upper+1; ++k) DD[k] = MININT;
        for (k=lower-1; k<=upper; ++k) II[k] = MININT;
        if (type1 == SUBSTITUTE) SS[start] = snake(start, i1, i2, j2);
        else if (type1 == DELETE) {
                DD[start] = i1;
                SS[start] = snake(start,i1,i2,j2);
        } else {
                II[start] = i1;
                SS[start] = snake(start,i1,i2,j2);
        }

        for (c=1; c<=midc; ++c) {
                t = max(lower, start-c);
                t1 = II[t-1];
                for (k=t; k<=min(upper, start+c); ++k) {
                        t2 = II[k];
                        II[k] = max(t1, SS[k]);
                        t1 = t2;
                        DD[k] = max(DD[k+1]+1, SS[k]);
                        SS[k] = snake(k, min(j2-k,max(max(SS[k]+1, II[k]), DD[k])), i2, j2);
                }
        }

        /* Allocate space for backward vectors */
        RS = (int *)ckalloc((rupper-rlower+1)*sizeof(int)) - rlower;
        RD = (int *)ckalloc((rupper-rlower+2)*sizeof(int)) - rlower + 1;
        RI = (int *)ckalloc((rupper-rlower+2)*sizeof(int)) - rlower;

        /* Backward computation */
        maxint = i2 + dist + max(M,N);
        for (k=rlower; k<=rupper; ++k) RS[k] = maxint;
        for (k=rlower-1; k<=rupper; ++k) RD[k] = maxint;
        for (k=rlower; k<=rupper+1; ++k) RI[k] = maxint;
        if (type2 == SUBSTITUTE)
           RI[rstart] = RD[rstart] = RS[rstart] = rsnake(rstart, i2, i1, j1, M);
        else if (type2 == DELETE) RD[rstart] = i2;
        else RI[rstart] = i2;

        for (c=1; c<=rmidc; ++c) {
                t = max(rlower, rstart-c);
                t1 = RD[t-1];
                for (k=t; k<=min(rupper, rstart+c); ++k) {
                        RS[k] = rsnake(k, max(j1-k, min(min(RS[k]-1,RD[k]),RI[k])),i1,j1,M);
                        t2 = RD[k];
                        RD[k] = min(t1-1, RS[k]);
                        t1 = t2;
                        RI[k] = min(RI[k+1], RS[k]);
                }
        }

        /* Find (mi, mj, mtype) such that
           the distance from (i1, j1, type1) to (mi, mj, mtype) is midc
           and the distance from (mi, mj, mtype) to (i2, j2, type2) is rmidc.
        */

        flag = FALSE;
        /* initialized for LINTing */
        mi = i1; mj = j1; mtype = type1;
        for (k=max(lower,rlower); k<=min(upper,rupper);++k) {

                if (SS[k]>=RS[k] || DD[k]>=RD[k] || II[k]>=RI[k]) {
                        if (DD[k]>=RD[k]) {
                                mi = DD[k];
                                mj = k+mi;
                                mtype = DELETE;
                        } else if (II[k] >= RI[k]) {
                                mi = II[k];
                                mj = k+mi;
                                mtype = INSERT;
                        } else {
                                mi = SS[k];
                                mj = k+mi;
                                mtype = SUBSTITUTE;
                        }

                        flag = TRUE;
                        break;
                }
        }
        /* Free working vectors */
        free(SS+lower);
        free(DD+lower);
        free(II+lower-1);
        free(RS+rlower);
        free(RD+rlower-1);
        free(RI+rlower);

        if (flag) {
                /* Find a path from (i1,j1,type1) to (mi,mj,mtype) */
                path(i1,j1,type1,mi,mj,mtype,midc,&head1,&tail1);

                /* Find a path from (mi,mj,mtype) to (i2,j2,type2) */
                path(mi,mj,mtype,i2,j2,type2,rmidc,&head2,&tail2);

                /* Join these two paths together */
                if (head1) tail1->next = head2;
                else head1 = head2;
        } else {
                (void)printf("Something wrong when dividing\n");
                head1 = NULL;
        }
        *head = head1;
        if (head2) *tail = tail2;
        else *tail = tail1;
}


int get_dist(int i1, int j1, int i2, int j2, int limit)
{
        int *SS, *DD, *II;
        int goal_diag;
        int c, k, t1, t2, t;
        int start, lower, upper;

        /* Compute the boundary diagonals */
        start = j1 - i1;
        lower = max(j1-i2, start-limit);
        upper = min(j2-i1, start+limit);
        goal_diag = j2-i2;

        if (goal_diag > upper || goal_diag < lower) {
                (void)printf("The two sequences are not really similar.\n");
                (void)printf("Please try an exact aligning method.\n");
                exit(1);
        }

        /* Allocate space for forward vectors */
        SS = (int *)ckalloc((upper-lower+1)*sizeof(int)) - lower;
        DD = (int *)ckalloc((upper-lower+2)*sizeof(int)) - lower;
        II = (int *)ckalloc((upper-lower+2)*sizeof(int)) - lower + 1;

        /* Initialization */
        for (k=lower; k<=upper; ++k) SS[k] = MININT;
        for (k=lower; k<=upper+1; ++k) DD[k] = MININT;
        for (k=lower-1; k<=upper; ++k) II[k] = MININT;
        SS[start] = snake(start, i1, i2, j2);

        if (SS[goal_diag] >= i2) {

                /* Free working vectors */
                free(SS+lower);
                free(DD+lower);
                free(II+lower-1);
                return 0;
        }

        for (c=1; c<=limit; ++c) {
                t = max(lower, start-c);
                t1 = II[t-1];
                for (k=t; k<=min(upper, start+c); ++k) {
                        t2 = II[k];
                        II[k] = max(t1, SS[k]);
                        t1 = t2;
                        DD[k] = max(DD[k+1]+1, SS[k]);
                        SS[k] = snake(k, min(j2-k,max(max(SS[k]+1, II[k]), DD[k])), i2, j2);

                }

                if (SS[goal_diag] >= i2) {
#ifdef STATS
                        (void)printf("get_dist = %d\n",c);
#endif

                        /* Free working vectors */
                        free(SS+lower);
                        free(DD+lower);
                        free(II+lower-1);
                        return c;
                }
        }

        /* Ran out of distance limit */
        return -1;
}


/* Condense_script - merge contiguous operations of the same type together */
void Condense_script(edit_script *head)
{
        edit_script *tp, *tp1;

        tp = head;
        while (tp != NULL) {
           while (((tp1 = tp->next) != NULL) && (tp->op_type == tp1->op_type)) {
                 tp->num = tp->num + tp1->num;
                 tp->next = tp1->next;
                 free(tp1);
           }
           tp = tp->next;
        }
}


/* reverse the edit script */
#ifdef AUXUTILS
void Reverse_Script(edit_script *head)
{
        edit_script *tp;

        tp = head;
        while (tp != NULL) {
           if (tp->op_type == INSERT) tp->op_type = DELETE;
           else if (tp->op_type == DELETE) tp->op_type = INSERT;
           tp = tp->next;
        }
}
#endif


#ifdef AUXUTILS
void Print_Script(edit_script *head, int M, int N)
{
        edit_script *tp;
        int i, j, k, count;

        i = j = 0;
        tp = head;

        while (tp != NULL) {
           if (tp->op_type == SUBSTITUTE) {
                k = 0;
                while (k < tp->num) {
                        count = 0;
                        while ((seq1[i] == seq2[j]) &&
                               (k<tp->num)) {
                                ++i;
                                ++j;
                                ++count;
                                ++k;
                        }
                        if (count > 0) (void)printf("copy %d\n", count);
                        if (k < tp->num) {
                                (void)printf("replace %c by %c\n", seq1[i], seq2[j]);
                                ++i;
                                ++j;
                                ++k;
                        }
                }
                /*
                if (tp->num > 1) (void)printf("%d substitutions\n", tp->num);
                else (void)printf("%d substitution\n", tp->num);
                */
           } else if (tp->op_type == INSERT) {
                if ((tp==head || tp->next==NULL) && (M <= N))
                        (void)printf("skip (second sequence) %d\n", tp->num);
                else {
                        /*
                        (void)printf("insert %d\n", tp->num);
                        */
                        (void)printf("insert ");
                        if (tp->num<=10)
                            for (k=j; k<j+tp->num; ++k) (void)printf("%c", seq2[k]);
                        else (void)printf(" %d ",tp->num);
                        (void)printf("\n");
                }
                j += tp->num;
           } else {     /* DEL */
                if ((tp==head || tp->next==NULL) && (M > N))
                        (void)printf("skip (first sequence) %d\n", tp->num);
                else {
                        /*
                        (void)printf("delete %d\n", tp->num);
                        */
                        (void)printf("delete ");
                        if (tp->num <= 10)
                            for (k=i; k<i+tp->num; ++k)
                                 (void)printf("%c", seq1[k]);
                        else
                            (void)printf("%d ",tp->num);
                        (void)printf("\n");
                }
                i += tp->num;
           }
           tp = tp->next;
        }
}
#endif

void S2A(edit_script *head, int *S, int flag)
{
        edit_script *tp;
        int *lastS, i;

        tp = head;
        lastS = S;
        while (tp != NULL) {
/*
        (void)printf("tp->op_type=%d, tp->num=%d\n",tp->op_type, tp->num);
*/
           if (tp->op_type == SUBSTITUTE) {
                for (i=0; i<tp->num; ++i) *lastS++ = 0;
           } else if (tp->op_type == INSERT) {
                *lastS++ = (!flag) ? tp->num : (0-tp->num);
           } else {     /* DELETE */
                *lastS++ = (!flag) ? (0 - tp->num) : tp->num;
           }
           tp = tp->next;
        }
        *(S-1) = lastS - S;
}


/* Alignment display routine */

static uchar ALINE[51], BLINE[51], CLINE[51];

#ifdef AUXUTILS
void DISPLAY(uchar A[], uchar B[], int M, int N, int S[], int AP, int BP)
{ 
  register uchar *a, *b, *c;
  register int    i,  j, op;
           int   lines, ap, bp;

  i = j = op = lines = 0;
  ap = AP;
  bp = BP;
  a = ALINE;
  b = BLINE;
  c = CLINE;
  while (i < M || j < N)
    { if (op == 0 && *S == 0)
        { op = *S++;
          *a = A[++i];
          *b = B[++j];
          *c++ = (*a++ == *b++) ? '|' : ' ';
        }
      else
        { if (op == 0)
            op = *S++;
          if (op > 0)
            { *a++ = ' ';
              *b++ = B[++j];
              op--;
            }
          else
            { *a++ = A[++i];
              *b++ = ' ';
              op++;
            }
          *c++ = '-';
        }
      if ((a >= ALINE+50) || (i >= M) && (j >= N))
        { *a = *b = *c = '\0';
          (void)printf("\n%6d ",50*lines++);
          for (b = ALINE+10; b <= a; b += 10)
            (void)printf("    .    :");
          if (b <= a+5)
            (void)printf("    .");
          (void)printf("\n%6d %s\n       %s\n%6d %s\n",ap,ALINE,CLINE,bp,BLINE);
          ap = AP + i;
          bp = BP + j;
          a = ALINE;
          b = BLINE;
          c = CLINE;
        }
    }
}
#endif

void CDISPLAY(uchar A[], uchar B[], int M, int N, int S[], int AP, int BP, int intron_strand)
{
  register uchar *a, *b, *c, sign;
  register int    i,  j, op, prevop, index, GTAG_score, CTAC_score;
           int   lines, ap, bp;

  i = j = op = prevop = lines = index = 0;
  sign = '*';
  ap = AP;
  bp = BP;
  a = ALINE;
  b = BLINE;
  c = CLINE;
  while (i < M || j < N)
    { if (op == 0 && *S == 0)
        {
          prevop = op;
          op = *S++;
          *a = A[++i];
          *b = B[++j];
          *c++ = (*a++ == *b++) ? '|' : ' ';
        }
      else
        { if (op == 0) {
            op = *S++;
          }
          if (op > 0)
            {
              if (intron_strand==2) {
                  prevop = op;
                  *a++ = ' ';
                  *b++ = B[++j];
                  *c++ = '-';
                  op--;
              } else {
                  if (prevop<=0) {
                    if (op>=MIN_INTRON) {
                      /* find intron orientation */
                      GTAG_score = CTAC_score = 0; 
                      
                      if (B[j+1]=='G') GTAG_score++;
                      else if (B[j+1]=='C') CTAC_score++;
                      if (B[j+2]=='T') 
                         { GTAG_score++; CTAC_score++; }
                      if (B[j+op-1]=='A')
                         { GTAG_score++; CTAC_score++; }
                      if (B[j+op]=='G') GTAG_score++; 
                      else if (B[j+op]=='C') CTAC_score++; 
                      
                      if (GTAG_score > CTAC_score) sign = '>';
                      else if (GTAG_score < CTAC_score) sign = '<';
                      else sign = '=';
                      
                      index = 1;
                      *c++ = sign;
                    } else {
                      index = 0;
                      *c++ = '-';
                    } 
                    prevop = op;
                    *a++ = ' ';
                    *b++ = B[++j];
                    op--;
                  } else {
                    /* not the first deletion in the group */
                    if (!index) {
                       prevop = op;
                      *a++ = ' ';
                      *b++ = B[++j];
                      *c++ = '-';
                       op--;
                    } else {
                        prevop = op;
                        switch (index) {
                           case 0:
                           case 1:
                           case 2: *a++ = ' '; *b++ = B[++j]; 
                                   *c++ = sign; op--; break;                    
                           case 3:
                           case 4: *a++ = ' '; *b++ = '.'; *c++ = '.';
                                    j++; op--; break;                   
                           case 5: *a++ = ' '; *b++ = '.'; *c++ = '.';
                                    j+= op-3; op = 3; break;
                           case 6:  
                           case 7:  
                           case 8: *a++ = ' '; *b++ = B[++j]; 
                                   *c++ = sign; op--; break;                    
                        }  
                        index++;
                    }   
                  }     
              }
            }
          else
            { 
              if (intron_strand==2) {
                  if (prevop>=0) {
                    if (-op>=MIN_INTRON) {
                      /* find intron orientation */
                      GTAG_score = CTAC_score = 0;
                
                      if (A[i+1]=='G') GTAG_score++;
                      else if (A[i+1]=='C') CTAC_score++;
                      if (A[i+2]=='T')
                         { GTAG_score++; CTAC_score++; }
                      if (A[i-op-1]=='A')
                         { GTAG_score++; CTAC_score++; }
                      if (A[i-op]=='G') GTAG_score++;
                      else if (A[i-op]=='C') CTAC_score++;
                    
                      if (GTAG_score > CTAC_score) sign = '>';
                      else if (GTAG_score < CTAC_score) sign = '<';
                      else sign = '=';

                      index = 1;
                      *c++ = sign;
                    } else  {
                      index = 0;
                      *c++ = '-';
                    }
                    prevop = op;
                    *a++ = A[++i];
                    *b++ = ' ';
                    op++;
                  } else {
                    /* not the first deletion in the group */
                    if (!index) {
                       prevop = op;
                      *a++ = A[++i];
                      *b++ = ' ';
                      *c++ = '-';
                      op++;
                    } else {
                        prevop = op;
                        switch (index) {
                           case 0:
                           case 1:
                           case 2: *a++ = A[++i]; *b++ = ' '; *c++ = sign; op++; break;
                           case 3:
                           case 4: *a++ = '.'; *b++ = ' '; *c++ = '.'; i++; op++; break;
                           case 5: *a++ = '.'; *b++ = ' '; *c++ = '.';
                                    i+= (-op)-3; op = -3; break;
                           case 6:
                           case 7:
                           case 8: *a++ = A[++i]; *b++ = ' '; *c++ = sign; op++; break;
                        }
                        index++;
                    }
                  }
              } else {
                prevop = op;
                *a++ = A[++i];
                *b++ = ' ';
                op++;
                *c++ = '-';
              }
            }
/*       *c++ = '-';    */
        }
      if ((a >= ALINE+50) || ((i >= M) && (j >= N)))
        { *a = *b = *c = '\0';
          (void)printf("\n%7d ",50*lines++);
          for (b = ALINE+10; b <= a; b += 10)
            (void)printf("    .    :");
          if (b <= a+5)
            (void)printf("    .");
          (void)printf("\n%7d %s\n        %s\n%7d %s\n",ap,ALINE,CLINE,bp,BLINE);
          ap = AP + i;
          bp = BP + j;
          a = ALINE;
          b = BLINE;
          c = CLINE;
        }
    }
}  


void Script_free(edit_script *head)
{
        edit_script *tp, *tp1;

        tp = head;
        while (tp != NULL) {
           tp1 = tp->next;
           free(tp);
           tp = tp1;
        }
}


static int snake(int k, int x, int endx, int endy)
{
        int y;

        if (x<0) return x;
        y = x+k;
        while (x<endx && y<endy && seq1[x]==seq2[y]) {
                ++x; ++y;
        }
        return x;
}


static int rsnake(int k, int x, int startx, int starty, int M)
{
        int y;

        if (x>M) return x;
        if ((startx<0) || (starty<0))
            (void)printf("TROUBLE!!! startx:  %5d,  starty:  %5d\n",startx, starty);
        if ((x>M) || (x+k>N))
             (void)printf("TROUBLE!!! x:  %5d,  y:  %5d\n",x,x+k);
        
        y = x+k;
        while (x>startx && y>starty && seq1[x-1]==seq2[y-1]) {
                --x; --y;
        }
        return x;
}


