/*
 * sim4.c - A program to align a cDNA sequence with a genomic sequence
 *          for the gene.
 */

#ifndef __lint
/*@unused@*/
static const char rcsid[] = 
"$Id: sim4b1.c,v 1.22 1998/12/13 21:48:54 florea Exp florea $";
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "sim4.h"
#include "simlib.h"
#include "sim4b1.h"
#include "Xtend1.h"
#include "scriptlib.h"

#define  EXTEND_FW (acc_flag?Xextend_fw:extend_fw)
#define  EXTEND_BW (acc_flag?Xextend_bw:extend_bw)

uchar      *seq1, *seq2;
int         M, N;
bool        ali_flag, acc_flag, rcflag, reverse;
int         file_type;
coords      last_GT, last_CT, last_AG, last_AC;

int       DRANGE, weight;
int       cutoff, mult;
int       icoverage, internal;
float     fcoverage;
#ifdef _STATS
float     marginals;
#endif


static int         numMSPs, K, W, X;
static int         G_score, C_score;
static int       *diag_lev;
static Msp_ptr     msp_list, *msp;
static Exon_ptr    exon_list;

static void   merge(Exon **,Exon **); 
static void   free_2D_Int(int **,int);
static void   free_this_script (struct edit ***,int);
static void   slide(int,char *,int *,int *,int *,int *, int *);
static Exon  *bmatch(uchar *,uchar *,int,int,int,int);
static Exon  *fmatch(uchar *,uchar *,int,int,int,int);
static void   compact_list(Exon **Lblock, Exon **Rblock);
static int    greedy(uchar *,uchar *,int,int,int,int,Exon **, Exon **);
static int    extend_bw(uchar *,uchar *,int,int,int,int,int *,int *);
static int    extend_fw(uchar *,uchar *,int,int,int,int,int *,int *);
static void   bput_scr(struct edit *,int,int,int,int,int,int,Exon **,Exon **);
static void   fput_scr(struct edit *,int,int,int,int,Exon **,Exon **);
static void   pluri_align(int *,Exon *,struct edit_script_list **);
static void   flip_list(Exon **,Exon **);
static void   add_word(int,int);
static void   extend_hit(int,int,const uchar *const,const uchar * const,int,int,int);
static void   sort_msps(void);
static void   heapify(int,int);
static int    smaller(int,int);
static void   search(uchar *,uchar *,int,int,int);
static int    link_msps(Msp_ptr *msp,int);
static void   msp2exons(Msp_ptr *, int);
static void   exon_cores(uchar*,uchar*,int,int,int,int,int,int,double,int);


/*  Not currently used: */
#ifdef AUXUTILS 
   static void   align(int *,Exon *,struct edit_script **); 
   static void   find_introns(Exon *, Intron **);
   static void   print_introns(Intron *);
#endif

/* seq1 = genomic  DNA (text); seq2 = cDNA */
struct edit_script_list *SIM4(uchar *in_seq1, uchar *in_seq2, 
                         int in_M, int in_N,
                         int in_W, int in_X, double in_F,
                         int *dist_ptr, Exon **Exons)
{
  int     cflag, diff, cost, rollbflag;
  int    u, v, I, J;
  Exon   *Lblock, *Rblock=NULL, *tmp_block, *last, *tmp_block1,
         *tmp_Lblock=NULL, *tmp_Rblock=NULL, *new;

  struct edit_script_list *head, *Script_head;
  struct edit_script *tail;
      uchar tmp[50];
      coords *sig;

  seq1 = in_seq1;
  seq2 = in_seq2;
  M = in_M;
  N = in_N;
  W = in_W;
  X = in_X;

  if (acc_flag) {
      last_AG.pos1 = last_AG.pos2 = last_AC.pos1 = last_AC.pos2 = 0;
      last_GT.pos1 = last_GT.pos2 = last_CT.pos1 = last_CT.pos2 = 0;
  }

  /* Compute the distance between two sequences A and B */

  *dist_ptr = 0;

  exon_cores(seq1-1, seq2-1, M, N, 1, 1, 0, W,in_F, PERM); 

  tmp_block = Lblock = exon_list;
  while (tmp_block) {
     if (tmp_block->next_exon==NULL) 
         Rblock = tmp_block;
     tmp_block = tmp_block->next_exon; 
  }

  /* enclose the current path in the (0,0,0,0) and (M+1,N+1,0,0) brackets */

  new = (Exon *)ckalloc(sizeof(Exon));
  new->to1 = 0;
  new->to2 = 0;
  new->from1 = 0;
  new->from2 = 0;
  new->length = 0;
  new->flag = (char)0;
  new->next_exon = Lblock;
  Lblock = new;

  if (Rblock == NULL)
      Rblock = Lblock;

  new = (Exon *)ckalloc(sizeof(Exon));
  new->from1 = M+1;
  new->from2 = N+1;
  new->to1 = 0;
  new->to2 = 0;
  new->length = 0;
  new->flag = (char)0;
  new->next_exon = NULL;
  Rblock->next_exon = new;

  #ifdef DEBUG
  (void)fprintf(stderr,"\n====================LSIS:\n\n");
  tmp_block = Lblock;
  while (tmp_block!=NULL) {
          (void)fprintf(stderr," [ %d, %d, %d, %d, l: %d ]\n ",
                        tmp_block->from1, tmp_block->from2,
                        tmp_block->to1, tmp_block->to2,
                         tmp_block->length);
          tmp_block = tmp_block->next_exon;
  }
  #endif

  tmp_block = Lblock;
  while ((tmp_block1 = tmp_block->next_exon)!=NULL) {
     rollbflag = 0;
     diff = (int)(tmp_block1->from2-tmp_block->to2-1);
     if (diff) {
     if (diff<0) {
         int   best_u, l0, l1;
         int    GTAG_score, CTAC_score;
         uchar *s1, *s2;

         /* resolve overlap using the GT-AG criterion */
         /* u-1 = actual position in the sequence */

         l0 = tmp_block->length-diff;
         l1 = tmp_block1->length;

         best_u = u = tmp_block1->from2-1;  
         s1 = seq1+tmp_block->to1-(tmp_block->to2-u);
         s2 = seq1-2+tmp_block1->from1+u-tmp_block1->from2;

         cost = 0;
         while (s1<=seq1+tmp_block->to1) {
           GTAG_score = CTAC_score = 0;
           GTAG_score += ((char)(*s1)=='G') ? 1 : 0;
           GTAG_score += ((char)(*(s1+1))=='T') ? 1 : 0;
           GTAG_score += ((char)(*s2)=='A') ? 1 : 0;
           GTAG_score += ((char)(*(s2+1))=='G') ? 1 : 0;
                          
           if (GTAG_score > abs(cost) && ((l0>=8) || (l1>=8))) {
               cost = GTAG_score;
               best_u = u; 
               if (cost == 4) break;
           }

           CTAC_score += ((char)(*s1)=='C') ? 1 : 0;
           CTAC_score += ((char)(*(s1+1))=='T') ? 1 : 0;
           CTAC_score += ((char)(*s2)=='A') ? 1 : 0;
           CTAC_score += ((char)(*(s2+1))=='C') ? 1 : 0;
          
           if (CTAC_score > abs(cost)) {
               cost = -CTAC_score;
               best_u = u;
               if (cost == 4) break;
           }

           u++; s1++; s2++;
           l0++; l1--;
         }

         /* mark intron orientation according to the sign of cost NOTE */

         tmp_block1->from1 += best_u+1-tmp_block1->from2;
         tmp_block1->from2 = best_u+1;
         if (((l1=tmp_block1->to2-tmp_block1->from2+1)<=0) || (l1<8)) {
             /* remove exon associated with tmp_block1 */ 
             tmp_block->next_exon = tmp_block1->next_exon;
             free(tmp_block1);
             tmp_block1 = tmp_block;
         } 
             
         tmp_block->to1 -= tmp_block->to2-best_u;
         tmp_block->to2 = best_u;
         if (((l0=tmp_block->to2-tmp_block->from2+1)<=0) || (l0<8)) {
             
             /* remove exon defined by tmp_block */
             exon_list = Lblock;
             while ((exon_list!=NULL) && (exon_list->next_exon!=tmp_block))
                exon_list = exon_list->next_exon;
             if (exon_list==NULL) 
                 fatal("Consistency test failed. Check program.");
             exon_list->next_exon = tmp_block->next_exon; /* NEW */
             if (tmp_block->flag==(char)1)
                 exon_list->flag = (char)1;
             if (l0>0) rollbflag = 1;
             free(tmp_block);
             tmp_block = exon_list;
         } 
/*
         (void)printf("Intron signal cost: %d, [ %d : %d, %d : %d ]\n",
                      cost,
                      tmp_block->to1,tmp_block1->from1,
                      tmp_block->to2,tmp_block1->from2);
*/

         if (tmp_block->to1)
             tmp_block->length = tmp_block->to2-tmp_block->from2+1;
         if (tmp_block1->to1)
             tmp_block1->length = tmp_block1->to2-tmp_block1->from2+1;

     } else {
         /* bridge the gap */
         cflag = (tmp_block1->to2 && tmp_block->to2) ? 0 : 1;
         if (diff && (tmp_block1->from1-tmp_block->to1-1>0)) { 
             if (!cflag) {
              if (diff<=MAX_GRINIT) {
                cost = greedy(seq2+tmp_block->to2,
                            seq1+tmp_block->to1, 
                            diff,
                            tmp_block1->from1-tmp_block->to1-1,
                            tmp_block->to2,tmp_block->to1,
                            &tmp_Lblock, &tmp_Rblock);
                flip_list(&tmp_Lblock, &tmp_Rblock);
              } else cost = max(W,(int)(P*diff+1))+1;

              if (cost>max(W,(int)(P*diff+1))) {
                  if (!tmp_block->flag && !tmp_block1->flag) {
                      exon_cores(seq1+tmp_block->to1-1,
                                 seq2+tmp_block->to2-1,
                                 tmp_block1->from1-tmp_block->to1-1,
                                 diff,
                                 tmp_block->to1+1,
                                 tmp_block->to2+1,
                                 1,
                                 min(8,W),
                                 in_F,
                                 (min(8,W)==W) ? PERM : TEMP);
                      tmp_Lblock = tmp_Rblock = exon_list;
                      while ((tmp_Rblock!=NULL) && (tmp_Rblock->next_exon!=NULL)) 
                         tmp_Rblock = tmp_Rblock->next_exon;

                      if (tmp_Lblock) rollbflag = 1;
                      else rollbflag = 0;   /* already 0 */
                  } else 
                        tmp_Lblock = tmp_Rblock = NULL;
              }
             } else if (tmp_block1->to1) {
                   /* find last_AG, last_AC */
                   if (acc_flag) {
                     for (v=tmp_block1->from1-1; v<=tmp_block1->to1-3; v++) 
                        if (!strncmp((char *)(seq1+v-2),"AG",(size_t)2)) {
                          last_AG.pos1 = v+1;
                          last_AG.pos2 = tmp_block1->from2+
                                         (v-tmp_block1->from1)+1; 
                          break;
                        } 
                     for (v=tmp_block1->from1-1; v<=tmp_block1->to1-3; v++)
                        if (!strncmp((char *)(seq1+v-2),"AC",(size_t)2)) {
                          last_AC.pos1 = v+1;
                          last_AC.pos2 = tmp_block1->from2+
                                         (v-tmp_block1->from1)+1;
                          break;
                        }
                   } /* end acc_flag */       

                   diff = (int)(min(diff,(int)(MAX_GRINIT/2)));
                   u = min(4*diff,tmp_block1->from1-tmp_block->to1-1); 
                   cost = EXTEND_BW(seq2+tmp_block->to2+
                                    (tmp_block1->from2-tmp_block->to2-1)-diff,
                                    seq1+tmp_block->to1+
                                    (tmp_block1->from1-tmp_block->to1-1)-u,
                                    (int)diff, u,
                                    tmp_block->to2+
                                    (tmp_block1->from2-tmp_block->to2-1)-diff,
                                    tmp_block->to1+
                                    (tmp_block1->from1-tmp_block->to1-1)-u,
                                    &I, &J);
                   tmp_block1->from2 = I+1;
                   tmp_block1->from1 = J+1;
                   tmp_block1->length = tmp_block1->to2-tmp_block1->from2+1;

                   /* use blast if marginal gap still exists, and this is first scan */
                   if (!(diff=(int)(tmp_block1->from2-tmp_block->to2-1)) ||
                       tmp_block1->flag) {
                       /* blast-treated region or no gap */
                       tmp_Rblock = tmp_Lblock = NULL;
                   } else {
                      exon_cores(seq1+tmp_block->to1-1,
                                 seq2+tmp_block->to2-1,
                                 tmp_block1->from1-tmp_block->to1-1,
                                 diff,
                                 tmp_block->to1+1,
                                 tmp_block->to2+1,
                                 1,
                                 min(10,W),
                                 in_F,
                                 (min(10,W)==W) ? PERM : TEMP);
                      tmp_Lblock = tmp_Rblock = exon_list;
                      while (tmp_Rblock && tmp_Rblock->next_exon)
                         tmp_Rblock = tmp_Rblock->next_exon;

                      if (tmp_Lblock) rollbflag = 1;
                   }

             } else {
                   if (acc_flag) {
                     for (v=tmp_block->to1; v>=tmp_block->from1; v--)
                        if (!strncmp((char *)(seq1+v),"GT",(size_t)2)) {
                          last_GT.pos1 = v;
                          last_GT.pos2 = tmp_block->to2-(tmp_block->to1-v);
                          break;
                        }
                     for (v=tmp_block->to1; v>=tmp_block->from1; v--)
                        if (!strncmp((char *)(seq1+v),"CT",(size_t)2)) {
                          last_CT.pos1 = v;
                          last_CT.pos2 = tmp_block->to2-(tmp_block->to1-v);
                          break;
                        }
                   }

                   diff = (int)(min(diff,(int)(MAX_GRINIT/2)));
                   cost = EXTEND_FW(seq2+tmp_block->to2,
                                    seq1+tmp_block->to1,
                                    diff,
                                    min(4*diff,tmp_block1->from1-tmp_block->to1-1),
                                    tmp_block->to2,tmp_block->to1,
                                    &I, &J);
                   if (tmp_block->to1) {
                       tmp_block->to2 = I;
                       tmp_block->to1 = J;
                       tmp_block->length = tmp_block->to2-tmp_block->from2+1;
                       tmp_Rblock = tmp_Lblock = NULL;
                   } else 
                       /* special case: no initial exon */
                       tmp_Lblock = tmp_Rblock = NULL;
                    
                   /* use blast if marginal gap still exists, and this is first scan */
                   if (!(diff=(int)(tmp_block1->from2-tmp_block->to2-1)) ||
                        tmp_block->flag) {
                       /* blast-treated region or no gap */
                       tmp_Rblock = tmp_Lblock = NULL;
                   } else {
                      exon_cores(seq1+tmp_block->to1-1,
                                 seq2+tmp_block->to2-1,
                                 tmp_block1->from1-tmp_block->to1-1,
                                 diff,
                                 tmp_block->to1+1,
                                 tmp_block->to2+1,
                                 1,
                                 min(10,W),
                                 in_F,
                                 (min(10,W)==W) ? PERM : TEMP);
                      tmp_Lblock = tmp_Rblock = exon_list;
                      while (tmp_Rblock && tmp_Rblock->next_exon)
                         tmp_Rblock = tmp_Rblock->next_exon;

                      if (tmp_Lblock) rollbflag = 1;
                   }
             } 
           } else if (diff) {
                  tmp_Rblock = tmp_Lblock = NULL;
/* AICI 
                  tmp_Rblock = tmp_Lblock = (Exon *)ckalloc(sizeof(Exon));
                  tmp_Lblock->to1 = tmp_block1->from1-1;
                  tmp_Lblock->from1 = tmp_block->to1+1;
                  tmp_Lblock->from2 = tmp_block1->from2-1;
                  tmp_Lblock->to2 = tmp_block->to2+1;
                  tmp_Lblock->flag = 0;
                  tmp_Lblock->length = new->to2-new->from2+1;
                  tmp_Lblock->next_exon = NULL;
*/
           }
   
           /* merge block in the exon list; make connections 
              to the previous list of blocks; maintain 
              increasing order */

           if (tmp_Lblock) {       
               tmp_block->next_exon = tmp_Lblock;
               tmp_Rblock->next_exon = tmp_block1;
               merge(&tmp_block,&tmp_block1);
           }
     }
    }  /* diff!=0 */
    if (!rollbflag) tmp_block = tmp_block1;
  }

  /* just printing ... */
  #ifdef DEBUG
  (void)fprintf(stderr,"\n====================EXTENSIONS:\n\n");
  tmp_block = Lblock;
  while (tmp_block!=NULL) {
          (void)fprintf(stderr," [ %d, %d, %d, %d, l: %d ]\n",
                        tmp_block->from1, tmp_block->from2,
                        tmp_block->to1, tmp_block->to2,tmp_block->length);
          tmp_block = tmp_block->next_exon;
  }
  #endif

  /* compaction step; note: it resets the right end of the list to
     the last item in the block list
   */

  compact_list(&(Lblock->next_exon), &Rblock);

  /* just printing ... */
  #ifdef DEBUG
  (void)fprintf(stderr,"\n====================NORMALIZATION:\n\n");
  tmp_block = Lblock;
  while (tmp_block!=NULL) {
          (void)fprintf(stderr," [ %d, %d, %d, %d, l: %d ]\n",
                        tmp_block->from1, tmp_block->from2,
                        tmp_block->to1, tmp_block->to2,tmp_block->length);
          tmp_block = tmp_block->next_exon;
  }
  #endif


  /* eliminate marginal small blocks: begin sequence */
  /* resets the empty alignment to one block (Lblock) only */

  tmp_block = Lblock->next_exon;

  while ((tmp_block!=NULL) && (tmp_block->length<W) && tmp_block->to1) {
          tmp_block1 = tmp_block; /* free */
          tmp_block = tmp_block->next_exon;
          free(tmp_block1); /* free */
  }
  Lblock->next_exon = tmp_block;

  /* eliminate marginal small blocks: end sequence */

  last = Lblock->next_exon;
  tmp_block = last;
  while (tmp_block!=NULL) {
     if (tmp_block->length>=W)
         last = tmp_block;
     tmp_block = tmp_block->next_exon;
  }
  if (last && last->to1)
      last->next_exon = Rblock->next_exon;
  Rblock = last;

  /* slide exon boundaries for optimal intron signals */
   
  { /* start of block */ 
    int score;
 
    tmp_block = Lblock->next_exon;
    while (tmp_block && (tmp_block1=tmp_block->next_exon) &&
           tmp_block1->to1) {
       if (tmp_block1->from2-tmp_block->to2-1==0) {    
           slide(3,&tmp_block->ori,&tmp_block->to1,&tmp_block1->from1, 
                   &tmp_block->to2,&tmp_block1->from2, &score); 
           if (score>=0) G_score += score;
           else C_score += score;
           tmp_block->length = tmp_block->to2-tmp_block->from2+1;
           tmp_block1->length = tmp_block1->to2-tmp_block1->from2+1;
       } else
           tmp_block->ori = 'E';   
       tmp_block = tmp_block1; 
    }
  }  /* end of block */
   
  /* if high accuracy requirement, adjust boundaries of marginal exons */
  if (acc_flag) {
      tmp_block = Lblock->next_exon;

      /* condition for non-signal */
      if (tmp_block && tmp_block->to1 &&
          (strncmp((char *)(seq1+tmp_block->from1-3), END_SIG, (size_t)2) || 
          (tmp_block->from2!=1))) {
          sig = (G_score>=abs(C_score)) ? &last_AG : &last_AC; 
          if (sig->pos1 && (sig->pos2<=20)) {
              /* generated in extend_bw */
	      assert(sig->pos2 > 1);
              (void)strcpy((char *)tmp,END_SIG);
              (void)strncpy((char *)(tmp+2),(char *)seq2,(size_t)sig->pos2-1);
              (void)strcpy((char *)(tmp+sig->pos2+1), START_SIG);
              new = bmatch(seq1,tmp,tmp_block->from1-3,sig->pos2+3,1,1); 
              if (new) {
                  Lblock->next_exon->from1 = sig->pos1;
                  Lblock->next_exon->from2 = sig->pos2;
                  Lblock->next_exon->length -= sig->pos2-1;
                  new->next_exon = Lblock->next_exon;
                  new->ori = (G_score>=abs(C_score)) ? 'G' : 'C';
                  Lblock->next_exon = new; 
              }
          }
      }
      while (tmp_block && tmp_block->next_exon && tmp_block->next_exon->to1) 
             tmp_block = tmp_block->next_exon;
      if (tmp_block && tmp_block->to1 &&
          (strncmp((char *)(seq1+tmp_block->to1),START_SIG,(size_t)2) || (tmp_block->to2!=N))) {   
          sig = (G_score>=abs(C_score)) ? &last_GT : &last_CT;
          if (sig->pos1 && (N-sig->pos2<=20)) {
	      assert(N-sig->pos2 >= 0);
              (void)strcpy((char *)tmp,END_SIG);
              (void)strncpy((char *)(tmp+2),(char *)(seq2+sig->pos2),
			    (size_t)N-sig->pos2);
              (void)strcpy((char *)(tmp+N-sig->pos2+2),START_SIG);
              new = fmatch(seq1+sig->pos1-1,tmp,
                           M-sig->pos1+1,N-sig->pos2+4,
                           sig->pos1-1,sig->pos2+1);
              if (new) {
                  tmp_block->to1 = sig->pos1;
                  tmp_block->to2 = sig->pos2;
                  new->next_exon = tmp_block->next_exon;
                  tmp_block->next_exon = new;
                  tmp_block->ori = (G_score>=abs(C_score)) ? 'G' : 'C';
              }
          }
      } 
  }

#ifdef _STATS
  tmp_block = Lblock;
  if (!tmp_block->next_exon)
      /* no alignment found */ 
      marginals = 2.0;
  else {
    while (tmp_block) {
      if ((tmp_block1 = tmp_block->next_exon)!=NULL) {
       if (!tmp_block->to1 && tmp_block1->to1)
           marginals += (float)(tmp_block1->from2-1)/
                               tmp_block1->to2;
       else if (tmp_block->to1 && !tmp_block1->to1)
           marginals += (float)(N-tmp_block->to2)/
                               (N-tmp_block->from2+1);
       else if (!tmp_block->to1 && !tmp_block1->to1)
           marginals = 2.0;
      }   
      tmp_block = tmp_block->next_exon;
    }
  }
#endif

/* just printing ... */
   if (!ali_flag) {
       pluri_align(dist_ptr,Lblock,&Script_head);

       icoverage = 0;
       internal = 1;
       mult = 0;
       tmp_block = Lblock->next_exon;
       while (tmp_block!=NULL) {
          icoverage += tmp_block->length;
          if (tmp_block->length) mult++;
          tmp_block = tmp_block->next_exon;
       }
       fcoverage = ((float)icoverage)/N;
#ifdef _STATS
       marginals = marginals/2;
#endif
       tmp_block = Lblock;
       if ((tmp_block->next_exon==NULL) ||
           !tmp_block->next_exon->to1)
          internal = 0;
       while (tmp_block!=NULL) {  
          if (tmp_block->to1) {
             if (((tmp_block1=tmp_block->next_exon)!=NULL) && 
                 (tmp_block1->from2-tmp_block->to2-1>0) &&
                 tmp_block1->to1)
                        internal = 0;
          } 
          tmp_block = tmp_block->next_exon;
       }
       *Exons = Lblock->next_exon;
       free_align(Script_head);
       free(Lblock);
       return NULL;

   } else { 
      
       /* alignment output */
       flip_list(&Lblock,&Rblock); 

       icoverage = 0;   
       internal = 1;
       mult = 0;
       tmp_block = Lblock->next_exon;
       while (tmp_block!=NULL) {
          icoverage += tmp_block->length;
          if (tmp_block->length) mult++;
          tmp_block = tmp_block->next_exon;
       }
       fcoverage = ((float)icoverage)/N;
#ifdef _STATS
       marginals = marginals/2;
#endif
       tmp_block = Lblock;
       if ((tmp_block->next_exon==NULL) ||
           !tmp_block->next_exon->to1)
          internal = 0;
       while (tmp_block!=NULL) {
          if (tmp_block->to1) {
          if (((tmp_block1=tmp_block->next_exon)!=NULL) && 
              (tmp_block1->from2-tmp_block->to2-1>0) &&
              tmp_block1->to1)
                  internal = 0;
          }
          tmp_block = tmp_block->next_exon;
       }
     
       pluri_align(dist_ptr,Lblock,&Script_head);

       head = Script_head;
       while (head!=NULL) {

          /* Merge the contiguous same operations together */
          Condense_script(head->script);

          /* Reset tail (due to the condensation) */
          tail = head->script;
          while ( tail && (tail->next != NULL)) tail = tail->next;
          head = head->next_script;
       }
       flip_list(&Lblock,&Rblock);
       *Exons = Lblock->next_exon; 
       free(Lblock); 
       return Script_head;
   }
}


struct hash_node {
        int ecode;             /* integer encoding of the word */
        int pos;                /* positions where word hits query sequence
 */
        struct hash_node *link; /* next word with same last 7.5 letters */
};

#define HASH_SIZE 32767        /* 2**15 - 1 */
static struct hash_node *phashtab[HASH_SIZE+1];
static struct hash_node **hashtab;
static int encoding[NACHARS];
static int mask;
static int *next_pos, *pnext_pos;

/* --------------------   exon_cores()   --------------------- */

static void  exon_cores(uchar *s1, uchar *s2, int len1, int len2, int offset1, int offset2, int flag, int in_W, double F, int type)
{
    int       i, W, last_msp, lower, upper;
    int     *allocated;
    Exon     *tmp_block;


    /* compute expected length of longest exact match .. */
    K = (int) log(.75*len1*len2)/log(4.0);
    /* .. and throw in a fudge factor */
    K *= F;

    numMSPs = 0;
    exon_list = NULL;

    allocated = ckalloc((len1+len2+1)*sizeof(int));
    lower = ((file_type==1) || (file_type==2 && type==TEMP)) ? -len1 : -len2;   
    upper = ((file_type==1) || (file_type==2 && type==TEMP)) ?  len2 :  len1;     
    diag_lev = allocated - lower;
    for (i=lower; i<=upper; ++i) diag_lev[i]=0;

    W = min(in_W,len2);
    switch (file_type) {
       case 1:  bld_table(s2,len2,W, type);
                /* use longer sequence for permanent tables */
                search(s1,s2,len1,len2,W);
                break;
       case 2:  if (type!=TEMP) {
                    uchar *aux; int   auxi;

                    aux = s1; s1 = s2; s2 = aux; 
                    auxi = len1; len1 = len2; len2 = auxi;
                }
                bld_table(s2,len2,W, type);
                /* use longer sequence for permanent tables */
                search(s1,s2,len1,len2,W); 
                /* NOT NEEDED: local variables and not used anymore. */  
                if (type!=TEMP) {
                    /* uchar *aux;  */
                    register int   auxi; 
                    Msp_ptr mp; 

/*
                    aux = s1; s1 = s2; s2 = aux;
                    auxi = len1; len1 = len2; len2 = auxi;
*/
                    for (mp=msp_list, i=0; i<numMSPs; i++) {
			auxi = mp->pos1;
			mp->pos1 = mp->pos2;
			mp->pos2 = auxi;
			mp = mp->next_msp;
                    }
                }
                /* */
                break;
       default: fatal("Unexpected value for parameter C.");
    }
  
    free(allocated);
    if (type==TEMP) {
        register struct hash_node *hptr, *tptr;
        register int    hval;
        
        free(next_pos);
        for (hval=0; hval<HASH_SIZE+1; hval++) {
             hptr = hashtab[hval];
             while (hptr) {
                tptr = hptr;
                hptr = hptr->link;
                free(tptr);
             }
        }    
        free(hashtab);
    }   

    msp = (Msp_ptr *) ckalloc(numMSPs*sizeof(Msp_ptr));
    { Msp_ptr mp = msp_list;
      for (i = 0; i < numMSPs; ++i) {
                msp[i] = mp;
		mp = mp->next_msp;
      }
    }

    sort_msps();    /* sort in order of mp->pos2, in the shorter seq */

    /* organize Blast hits (MSPs) into exons */
    last_msp = link_msps(msp, numMSPs);

#ifdef DEBUG
    for (i = last_msp; i >= 0; i = msp[i]->prev)
         (void)printf("%d-%d\n", msp[i]->pos1, msp[i]->pos1 + msp[i]->len - 1);
#endif

    msp2exons(msp,last_msp);

    /* now free msp[]? */
    for (i=0; i<numMSPs; ++i)
             free(msp[i]);
    free(msp);

    tmp_block = exon_list;
    while (tmp_block!=NULL) {
       tmp_block->length = tmp_block->to2-tmp_block->from2+1;
       tmp_block->to1 += offset1;
       tmp_block->from1 += offset1;
       tmp_block->to2 += offset2;
       tmp_block->from2 += offset2;
       tmp_block->flag = flag;

       tmp_block = tmp_block->next_exon;
    }

    return ;
}


static int link_msps(Msp_ptr *msp, int numMSPs)
{
	int i, j, f1, f2, best, diag, diff_diag, best_sc, try;
	

	for (best = -1, best_sc = MININT, i = 0; i < numMSPs; ++i) {
		f1 = msp[i]->pos1;      /* start position in seq1 */
		f2 = msp[i]->pos2;      /* start position in seq2 */
		diag = f1 - f2;
		msp[i]->prev = -1;
		msp[i]->Score = 0;
		for (j = 0; j < i; ++j) {
			diff_diag = diag - msp[j]->pos1 + msp[j]->pos2;
			if (diff_diag < -DRANGE ||
                           (diff_diag > DRANGE && diff_diag < MIN_INTRON) ||
                           (msp[j]->pos2+msp[j]->len-1-f2>L))

				continue;
			try = msp[j]->Score - abs(diff_diag);
			if (try > msp[i]->Score) {
				msp[i]->Score = try;
				msp[i]->prev = j;
			}
		}
		msp[i]->Score += (weight*msp[i]->score);
		if (msp[i]->Score > best_sc) {
			best = i;
			best_sc = msp[i]->Score;
		}
	}

	return best;
}

/* -----------   build table of W-tuples in one of the sequences  ------------*/

void bld_table(uchar *s,int len, int in_W, int type)
{
        int ecode;
        int i;
        uchar *t;

        if (type == PERM) {
            mask = (1 << (in_W+in_W-2)) - 1;
            next_pos = pnext_pos;
            hashtab = phashtab;  
            return; 
        }   

        /* perform initializations */
        if (type == INIT) {
 
           /* perform initializations */
           for (i=0; i<NACHARS; ++i) encoding[i]=0;

           encoding['C'] = 1;
           encoding['G'] = 2;
           encoding['T'] = 3;
           mask = (1 << (in_W+in_W-2)) - 1;
           /* added +1 in the allocation below */
           next_pos = pnext_pos = (int *)ckalloc((len+1)*sizeof(int));
           hashtab = phashtab;
        } else {
           mask = (1 << (in_W+in_W-2)) - 1;
           next_pos = (int *)ckalloc((len+1)*sizeof(int));
           hashtab = (struct hash_node **)
                      ckalloc((HASH_SIZE+1)*sizeof(struct hash_node *));
           for (i=0; i<=HASH_SIZE; ++i) hashtab[i] = NULL;
        }
    
        t = s;
        ecode = 0L;
        for (i = 1; i < in_W; ++i)
                ecode = (ecode << 2) + encoding[*++t];
        for (; (i<=len) && (*++t); i++) {
                ecode = ((ecode & mask) << 2) + encoding[*t];
                add_word(ecode, (int)(t - s));
        }
}

/* add_word - add a word to the table of critical words */
static void add_word(int ecode, int pos)
{
        struct hash_node *h;
        int hval;

        hval = ecode & HASH_SIZE;
        for (h = hashtab[hval]; h; h = h->link)
                if (h->ecode == ecode)
                        break;
        if (h == NULL) {
                h = (struct hash_node *) ckalloc (sizeof(struct hash_node));
                h->link = hashtab[hval];
                hashtab[hval] = h;
                h->ecode = ecode;
                h->pos = -1;
        }
        next_pos[pos] = h->pos;
        h->pos = pos;
}

/* -----------------------   search the other sequence   ---------------------*/

static void search(uchar *s1, uchar *s2, int len1, int len2, int in_W)
{
        register struct hash_node *h;
        register uchar *t;
        register int ecode, hval;
        int i;
        int p;

        t = s1;
        ecode = 0L;
        for (i = 1; i < in_W; ++i)
                ecode = (ecode << 2) + encoding[*++t];
        for (; (i<=len1) && (*++t); i++) {
                ecode = ((ecode & mask) << 2) + encoding[*t];
                hval = ecode & HASH_SIZE;
                for (h = hashtab[hval]; h; h = h->link)
                        if (h->ecode == ecode) {
                                for (p = h->pos; p >= 0; p = next_pos[p])
                                        extend_hit((int)(t-s1),p,s1,s2,len1,len2, in_W);
                                break;
                        }
        }
}

/* extend_hit - extend a word-sized hit to a longer match */
static void extend_hit(int pos1, int pos2, 
		       const uchar * const s1, const uchar * const s2,
		       int len1, int len2, int in_W)
{
        const uchar *beg2, *beg1, *end1, *q, *s;
        int right_sum, left_sum, sum, diag, score;

        diag = pos2 - pos1;
        if (diag_lev[diag] > pos1)
                return;

        /* extend to the right */
        left_sum = sum = 0;
	q = s1+1+pos1;
        s = s2+1+pos2;
        end1 = q;
        while ((*s != '\0') && (*q != '\0') &&
               (s<=s2+len2) && (q<=s1+len1) && sum >= left_sum - X) {
                sum += ((*s++ == *q++) ? MATCH : MISMATCH);
                if (sum > left_sum) {
                        left_sum = sum;
                        end1 = q;
                }
        }

        /* extend to the left */
        right_sum = sum = 0;
        beg1 = q = (s1+1+pos1) - in_W;
        beg2 = s = (s2+1+pos2) - in_W;
        while ((s>s2+1) && (q>s1+1) && sum >= right_sum - X) {
                sum += ((*(--s) == *(--q)) ? MATCH : MISMATCH);
                if (sum > right_sum) {
                        right_sum = sum;
                        beg2 = s;
                        beg1 = q;
                }
        }

        score = in_W + left_sum + right_sum;
        if (score >= K) {
                Msp_ptr mp = (Msp_ptr)ckalloc(sizeof(*mp));

                mp->len = end1 - beg1;
                mp->score = score;
                mp->pos1 = beg1 - (s1+1);
                mp->pos2 = beg2 - (s2+1);
                mp->next_msp = msp_list;
                msp_list = mp;
                ++numMSPs;
        }

        /*diag_lev[diag] = (end1 - (s1+1)) + in_W; */
        diag_lev[diag] = (end1 - s1) - 1 + in_W;
}

/*  ----------------------------   sort the MSPs  ----------------------------*/

/* sort_msps - order database sequence for printing */
static void sort_msps(void)
{
        int i;
        Msp_ptr mp;

        for (i = (numMSPs/2) - 1; i >= 0; --i)
                heapify(i, (int) numMSPs-1);
        for (i = numMSPs-1; i > 0; --i) {
                mp = msp[0];
                msp[0] = msp[i];
                msp[i] = mp;
                if (i > 1)
                        heapify(0, i-1);
        }
}

/* heapify - core procedure for heapsort */
static void heapify(i, last)
int i, last;
{
        int lim = (last-1)/2, left_son, small_son;
        Msp_ptr mp;

        while (i <= lim) {
                left_son = 2*i + 1;
                if (left_son == last)
                        small_son = left_son;
                else
                        small_son = smaller(left_son, left_son+1);
                if (smaller(i, small_son) == small_son) {
                        mp = msp[i];
                        msp[i] = msp[small_son];
                        msp[small_son] = mp;
                        i = small_son;
                } else
                        break;
        }
}
/* smaller - determine ordering relationship between two MSPs */
static int smaller(i, j)
int i, j;
{
	Msp_ptr ki = msp[i], kj = msp[j];
 
	return ((ki->pos2 >= kj->pos2) ? i : j);
}

/* ---------------------  organize MSPs into exons  ---------------------*/

static void msp2exons(Msp_ptr *msp, int last_msp)  
{
        Exon     *new;
        Msp_ptr   mp;

        exon_list = NULL; 
        if (last_msp<0) return;

        mp = msp[last_msp];
        new = (Exon *)ckalloc(sizeof(Exon)); 
        new->from1 = mp->pos1;
        new->from2 = mp->pos2;
        new->to1 = mp->pos1+mp->len-1;
        new->to2 = mp->pos2+mp->len-1;
        new->next_exon = exon_list;
        exon_list = new;

last_msp = mp->prev;

while (last_msp>=0) {
   mp = msp[last_msp]; 
   if ((abs((exon_list->from2-exon_list->from1)-(mp->pos2-mp->pos1))<=L) 
       && (exon_list->from2-(mp->pos2+mp->len-1)<MAX_INTERNAL_GAP)) {
               /* merge with previous exon */
               exon_list->to1 = max(exon_list->to1,mp->pos1+mp->len-1);
               exon_list->to2 = max(exon_list->to2,mp->pos2+mp->len-1);
               exon_list->from1 = min(exon_list->from1,mp->pos1);
               exon_list->from2 = min(exon_list->from2,mp->pos2);
           } else {
               /* new exon */
               new = (Exon *)ckalloc(sizeof(Exon));
               new->from1 = mp->pos1;
               new->from2 = mp->pos2;
               new->to1 = mp->pos1+mp->len-1;
               new->to2 = mp->pos2+mp->len-1;
               new->next_exon = exon_list;
               exon_list = new;
           }
           last_msp = mp->prev;
        } 
}

/* ----------------------  print endpoints of exons  --------------------*/

#ifdef AUXUTILS 
static void find_introns(Exon *eleft, Intron **Ilist)
{
  Exon   *tmp_exon, *tmp_exon1;
  Intron *new, *tail;
  int     GTAG_score, CTAC_score;

  *Ilist = tail = NULL;
  if (!eleft) fatal("Something wrong in the exon list.\n");

  tmp_exon = eleft->next_exon;
  while ((tmp_exon!=NULL) && (tmp_exon1=tmp_exon->next_exon) &&
         tmp_exon1->to1) {
     
     new = (Intron *)ckalloc(sizeof(Intron)); 
     new->from1 = tmp_exon->to1+1;
     new->to1 = tmp_exon1->from1-1;
     new->from2 = tmp_exon->to2;
     new->to2 = tmp_exon1->from2;
     new->length = new->to1-new->from1+1;
     new->next_intron = NULL;
     if (!tail) *Ilist = new;
     else tail->next_intron = new; 
     tail = new;

     /* find orientation */
     GTAG_score = CTAC_score = 0;
     if (*(seq1+new->from1-1)=='G') GTAG_score++;
     else 
       if (*(seq1+new->from1-1)=='C') CTAC_score++;

     if (*(seq1+new->from1-1)=='T')   
          { GTAG_score++;  CTAC_score++; }

     if (*(seq1+new->to1-1)=='A') 
          { GTAG_score++;  CTAC_score++; }

     if (*(seq1+new->to1-1)=='G')  GTAG_score++;
     else 
       if (*(seq1+new->to1-1)=='C') CTAC_score++;

     if  (GTAG_score>=CTAC_score) 
          new->orientation = '+';
     else new->orientation = 'c';

     tmp_exon = tmp_exon1;
  }
 
}
#endif 

void complement_exons(Exon **left)
{ 
  Exon  *tmp_block, *right;
  char   prev, ch;

  prev = 'U'; /* unknown; should trigger error */
  tmp_block = *left;
  if ((file_type==1) && (rcflag || reverse)) {
      while (tmp_block) {
         if (tmp_block->to1) {
             register int aux;
             
             if (tmp_block->next_exon && tmp_block->next_exon->to1) {
               ch = tmp_block->ori;
               tmp_block->ori = prev; 
               switch (ch) {
                 case  'C': prev = 'G'; break;
                 case  'G': prev = 'C'; break;
                 case  'N': prev = 'N'; break;
                 case  'E': prev = 'E'; break;
                 default  : fatal("Inconsistency. Check exon orientation at complementation.");
               }
             } else 
               tmp_block->ori = prev;
             aux = tmp_block->from1;
             tmp_block->from1 = M+1-tmp_block->to1;
             tmp_block->to1 = M+1-aux;
             aux = tmp_block->from2;
             tmp_block->from2 = N+1-tmp_block->to2;
             tmp_block->to2 = N+1-aux;
         }
         tmp_block = tmp_block->next_exon;
         if (tmp_block && tmp_block->to1) 
             right = tmp_block;
      }
      flip_list(left,&right);
  }
}

void print_exons(Exon *left)
{
  Exon  *tmp_block, *tmp_block1;

  tmp_block = left;
  while (tmp_block!=NULL) {
     if (tmp_block->to1) {
         
         if (file_type==1) 
            (void)fprintf(stdout,"%d-%d  (%d-%d)   %d%%",
                          tmp_block->from2, tmp_block->to2,
                          tmp_block->from1, tmp_block->to1,
                          tmp_block->match);
         else  
            /* file_type==2 */
            (void)fprintf(stdout,"%d-%d  (%d-%d)   %d%%", 
                           tmp_block->from1, tmp_block->to1,
                           tmp_block->from2, tmp_block->to2,
                           tmp_block->match);
         if (((tmp_block1=tmp_block->next_exon)!=NULL) && tmp_block1->to1)
              switch (tmp_block->ori) {
                 case 'C':  (void)fprintf(stdout," <-\n");
                            break;
                 case 'E':  (void)fprintf(stdout," ==\n");
                            break;
                 case 'G':  (void)fprintf(stdout," ->\n");
                            break;
                 case 'N':  (void)fprintf(stdout," --\n");
                            break;
                 default :  fatal("Inconsistency. Check exon orientations.");
              }  
     }        
     tmp_block = tmp_block->next_exon;
  }  
}    

#ifdef AUXUTILS
static void print_introns(Intron_ptr intron_list) {
  Intron_ptr ep=intron_list;

  (void)printf("\nIntrons {\n\n");
  while (ep!=NULL) {
     (void)printf("genome: %6ld - %-6ld    cDNA: %5ld - %-5ld,  l: %-5d,  o: %c \n",
                   ep->from1, ep->to1, ep->from2, ep->to2, 
                   ep->length,
                   ep->orientation);
     ep = ep->next_intron;
  }
  (void)printf("}\n\n");
}
#endif


#ifdef AUXUTILS
static void align(int *dist_ptr,Exon *lblock,struct edit_script **head)
{
   int    diff;
   Exon *tmp_block=lblock, *tmp_block1;

   struct edit_script *new, *left, *right;

   *head = NULL;
   *dist_ptr = 0;

   while (((tmp_block1=tmp_block->next_exon)!=NULL) && tmp_block1->to1) {
         diff = get_dist(tmp_block1->from1-1, tmp_block1->from2-1,
                         tmp_block1->to1, tmp_block1->to2, 1000);

         if (diff<0) {
             (void)printf("The two sequences are not really similar.\n");
             (void)printf("Please try an exact method.\n");
             exit(1);        
         }

#ifdef STATS
         if (diff>P*(tmp_block1->to2-tmp_block1->from2+1))
              (void)printf("Warning: Distance threshold on segment exceeded.\n");
#endif

         path(tmp_block1->from1-1, tmp_block1->from2-1, SUBSTITUTE,
              tmp_block1->to1, tmp_block1->to2, SUBSTITUTE,
              diff, &left, &right);
         *dist_ptr += diff;

         if ((diff=tmp_block->from2-tmp_block1->to2-1)!=0) {
              new = (edit_script *) ckalloc(sizeof(edit_script));
              new->op_type = INSERT;
              new->num = diff;
              new->next = *head;
              *head = new;
              *dist_ptr += diff;
         }

         if ((diff=tmp_block->from1-tmp_block1->to1-1)!=0) {
              new = (edit_script *) ckalloc(sizeof(edit_script));
              new->op_type = DELETE;
              new->num = diff;
              new->next = *head;
              *head = new;
         }
         right->next = *head;
         *head = left;
         tmp_block = tmp_block1;
  }
  if (tmp_block1!=NULL) {
    if ((diff=tmp_block->from1-tmp_block1->to1-1)!=0) {
                new = (edit_script *) ckalloc(sizeof(edit_script));
                new->op_type = DELETE;
                new->num = diff;
                new->next = *head;
                *head = new;
    }

    if ((diff=tmp_block->from2-tmp_block1->to2-1)!=0) {
                new = (edit_script *) ckalloc(sizeof(edit_script));
                new->op_type = INSERT;
                new->num = diff;
                new->next = *head;
                *head = new;
                *dist_ptr += diff;
    }
  }
}
#endif

static void pluri_align(int *dist_ptr,Exon *lblock,struct edit_script_list **Aligns)
{
   int    tmpi, di_count, i, end1, end2, diff, ali_dist;
   uchar *a, *b;
   Exon  *tmp_block=lblock, *tmp_block1;

   struct edit_script_list *enew;
   struct edit_script *head, *tmp_script, *new, *left, *right;

   head = NULL;
   *Aligns = NULL;
   *dist_ptr = ali_dist = 0;

   end1 = M; 
   end2 = N;

   while (((tmp_block1=tmp_block->next_exon)!=NULL) && tmp_block1->to1) {

         if ((diff=tmp_block->from2-tmp_block1->to2-1)!=0) {
           if (tmp_block->to1) {
              enew = (edit_script_list *)ckalloc(sizeof(edit_script_list));
              enew->next_script = *Aligns;
              *Aligns = enew;
              (*Aligns)->script = head;
              (*Aligns)->offset1 = tmp_block->from1;
              (*Aligns)->offset2 = tmp_block->from2;
              (*Aligns)->len1 = end1-(*Aligns)->offset1+1;
              (*Aligns)->len2 = end2-(*Aligns)->offset2+1;
              (*Aligns)->score = ali_dist;
              ali_dist = 0;
              head = NULL;
            }
            end1 = tmp_block1->to1;
            end2 = tmp_block1->to2;  

         } else if (((diff=tmp_block->from1-tmp_block1->to1-1)!=0) &&
                      tmp_block->to1) {
              new = (edit_script *) ckalloc(sizeof(edit_script));
              new->op_type = DELETE;
              new->num = diff;
              new->next = head;
              head = new;
         } else if (diff) 
              end1 = tmp_block1->to1;

         diff = get_dist(tmp_block1->from1-1, tmp_block1->from2-1,
                         tmp_block1->to1, tmp_block1->to2, 1000);

         if (diff<0) {
             (void)printf("The two sequences are not really similar.\n");
             (void)printf("Please try an exact method.\n");
             exit(1);        
         }

#ifdef STATS
         if (diff>P*(tmp_block1->to2-tmp_block1->from2+1))
              (void)printf("Warning: Distance threshold on segment exceeded.\n");
#endif

         path(tmp_block1->from1-1, tmp_block1->from2-1, SUBSTITUTE,
              tmp_block1->to1, tmp_block1->to2, SUBSTITUTE,
              diff, &left, &right);
         *dist_ptr += diff;
         ali_dist += diff;

         a = seq1+tmp_block1->from1-1;
         b = seq2+tmp_block1->from2-1;
         tmpi = di_count = 0;
         tmp_script = left;  
         while (tmp_script) {
            switch (tmp_script->op_type) {
                case  DELETE:
                      di_count += tmp_script->num;
                      tmpi += tmp_script->num;
                      a += tmp_script->num;
                      break;
                case  INSERT:
                      di_count += tmp_script->num;
                      tmpi += tmp_script->num;
                      b += tmp_script->num;
                      break;
                case  SUBSTITUTE:
                      for (i=0; i<tmp_script->num; ++i, ++a, ++b)
                        if (*a!=*b) tmpi++;
                      break;
            }         
            tmp_script = tmp_script->next;
         }  
         tmp_block1->match = (int)rint(100*(1-2*tmpi/
                             (float)(tmp_block1->to1-tmp_block1->from1+1+
                              tmp_block1->to2-tmp_block1->from2+1+di_count)));

         right->next = head;
         head = left;
         tmp_block = tmp_block1;
  }


   /* at the beginning of the sequences */
  if (tmp_block1!=NULL) {

    if ((diff=tmp_block->from2-tmp_block1->to2-1)!=0) {
         enew = (edit_script_list *)ckalloc(sizeof(edit_script_list));
         enew->next_script = *Aligns;
         *Aligns = enew;
         (*Aligns)->offset1 = tmp_block->from1;
         (*Aligns)->offset2 = tmp_block->from2;
         (*Aligns)->len1 = end1-(*Aligns)->offset1+1;
         (*Aligns)->len2 = end2-(*Aligns)->offset2+1;
         (*Aligns)->script = head; 
         (*Aligns)->score = ali_dist;
 
    } else {

         /* modified to cut introns at the beginning of the sequence */
         enew = (edit_script_list *)ckalloc(sizeof(edit_script_list));
         enew->next_script = *Aligns;
         *Aligns = enew;
         (*Aligns)->offset1 = tmp_block->from1;
         (*Aligns)->offset2 = 1;
         (*Aligns)->len1 = end1-(*Aligns)->offset1+1;
         (*Aligns)->len2 = end2-(*Aligns)->offset2+1;
         (*Aligns)->script = head;
         (*Aligns)->score = ali_dist;
    }

  }
}


static int  greedy(uchar *s1, uchar *s2, int m, int n, int offset1, int offset2, Exon **lblock, Exon **rblock)
{
  int     col,                    /* column number */
          d,                      /* current distance */
          k,                      /* current diagonal */
          max_d,                  /* bound on size of edit script */
          Cost,
          blower,flower,          /* boundaries for searching diagonals */
          bupper,fupper,
          row,                    /* row number */
          DELTA,                  /* n-m  */
          MAX_D,
          ORIGIN;
  int     back, forth;            /* backward and forward limits at exit */

  int     **blast_d, **flast_d;   /* rows containing the last d */
  int     *min_row, *min_diag,    /* min (b)/ max (f) row (and diagonal) */
          *max_row, *max_diag;    /* reached for cost d=0, ... m.  */

  struct edit ***bscript,         /* corresponding edit scripts */
              ***fscript;
  struct edit *new;
  Exon *tmp_lblock, *tmp_rblock;

  ValNodePtr  data_list = NULL, prev = NULL;

  DELTA = n-m;
/*max_d = MAX_D = m+1; */
  max_d = MAX_D = max(W,(int)(P*m+1)); 
   
  if (DELTA<0) { 

        *lblock = *rblock = (Exon *)ckalloc(sizeof(Exon));
        (*lblock)->from1 = offset2+1;
        (*lblock)->from2 = offset1+1;
        (*lblock)->to1 = offset2+n;
        (*lblock)->to2 = offset1+m;
        (*lblock)->flag = 0;
        (*lblock)->next_exon = NULL;
        (*lblock)->length = m;

        return n-m+(int)(P*m+1);
      }

  ORIGIN = MAX_D;
  for (row=m, col=n; row>0 && col>0 && (s1[row-1]==s2[col-1]); row--,col--)
	/*LINTED empty loop body*/;

  if (row == 0) {
       /* hit last row; stop search */
       *lblock = *rblock = (Exon *)ckalloc(sizeof(Exon));
       (*lblock)->from2 = offset1+1;
       (*lblock)->to2 = offset1+m;
       (*lblock)->from1 = offset2-m+n+1;
       (*lblock)->to1 = offset2+n;
       (*lblock)->flag = 0;
       (*lblock)->next_exon = NULL;
       (*lblock)->length = m;

       return 0;
  }


  blast_d = (int **)ckalloc((MAX_D+1)*sizeof(int *)); /* NEW */
  bscript = (struct edit ***)ckalloc((MAX_D+1)*sizeof(struct edit **));
  for (d=0; d<=MAX_D; d++) {
       blast_d[d] = (int *)ckalloc((MAX_D+n+1)*sizeof(int));
       bscript[d] = (struct edit **)ckalloc((MAX_D+n+1)*sizeof(struct edit *));
  }

  for (d=0; d<=MAX_D; d++)
       for (k=0; k<=MAX_D+n; ++k)
	    blast_d[d][k]=m+1;
  blast_d[0][ORIGIN+DELTA] = row;
  bscript[0][ORIGIN+DELTA] = NULL;

  blower = ORIGIN + DELTA - 1;
  bupper = ORIGIN + DELTA + 1;

  for (row=0; row<n && row<m && (s1[row]==s2[row]); row++)
        /*LINTED empty loop body*/; 

  if (row == m) {
       /* hit last row; stop search */

       *lblock = *rblock = (Exon *)ckalloc(sizeof(Exon));
       (*lblock)->from2 = offset1+1;
       (*lblock)->to2 = offset1+m;
       (*lblock)->from1 = offset2+1;
       (*lblock)->to1 = offset2+m;
       (*lblock)->flag = 0;
       (*lblock)->next_exon = NULL;
       (*lblock)->length = m;

       free_this_script (bscript, MAX_D+1);
       free_2D_Int(blast_d, MAX_D+1);
      
       return 0;
  }

  flast_d = (int **)ckalloc((MAX_D+1)*sizeof(int *));
  fscript = (struct edit ***)ckalloc((MAX_D+1)*sizeof(struct edit **));
  for (d=0; d<=MAX_D; d++) {
       flast_d[d] = (int *)ckalloc((MAX_D+n+1)*sizeof(int));
       fscript[d] = (struct edit **)ckalloc((MAX_D+n+1)*sizeof(struct edit *));
  }


  for (d=0; d<=MAX_D; d++)
       for (k=0; k<=MAX_D+n; ++k)
	   flast_d[d][k]=-1;
  flast_d[0][ORIGIN] = row;
  fscript[0][ORIGIN] = NULL;

  flower = ORIGIN - 1;
  fupper = ORIGIN + 1;

  max_row = (int *)ckalloc((MAX_D+1)*sizeof(int));
  min_row = (int *)ckalloc((MAX_D+1)*sizeof(int));
  max_diag = (int *)ckalloc((MAX_D+1)*sizeof(int));
  min_diag = (int *)ckalloc((MAX_D+1)*sizeof(int));

  for (d=1; d<=MAX_D; d++) {
       min_row[d] = m+1;
       max_row[d] = -1;
  }
  min_row[0] = blast_d[0][ORIGIN+DELTA];
  min_diag[0] = ORIGIN+DELTA;
  max_row[0] = flast_d[0][ORIGIN];
  max_diag[0] = ORIGIN;

  back = forth = 0;

  d = 1;
  while (d <= max_d) {

        /* for each relevant diagonal ... */
        for (k = blower; k <= bupper; k++) {
             /* get space for the next edit instruction */
             new = (struct edit *)ckalloc(sizeof(struct edit));
             link_to_data_list((Pointer)new, &data_list, &prev);

             /* find a d on diagonal k */
             if (k==-d+DELTA+ORIGIN) {

                      /* move left from the last d-1 on diagonal k+1 */
                      row = blast_d[d-1][k+1];
                      new->link = bscript[d-1][k+1];
                      new->op = INSERT;
             }
             else if (k==d+DELTA+ORIGIN) {

                      /* move up from the last d-1 on diagonal k-1 */
                      row = blast_d[d-1][k-1]-1;
                      new->link = bscript[d-1][k-1];
                      new->op = DELETE;
             } else if ((blast_d[d-1][k]<=blast_d[d-1][k+1]) &&
                        (blast_d[d-1][k]-1<=blast_d[d-1][k-1])) {

                      /* substitution */
                      row = blast_d[d-1][k]-1;
                      new->link = bscript[d-1][k];
                      new->op = SUBSTITUTE;

             } else if ((blast_d[d-1][k-1]<=blast_d[d-1][k+1]-1) &&
                        (blast_d[d-1][k-1]<=blast_d[d-1][k]-1)) {
                      /* move right from the last d-1 on diagonal k-1 */
                      row = blast_d[d-1][k-1]-1;
                      new->link = bscript[d-1][k-1];
                      new->op = DELETE;
             } else  {
                      /* move left from the last d-1 on diagonal k+1 */
                      row = blast_d[d-1][k+1];
                      new->link = bscript[d-1][k+1];
                      new->op = INSERT;
             }
             /* code common to the three cases */
             new->line1 = row;
             new->line2 = col = row + k - ORIGIN;
             bscript[d][k] = new;

             /* slide up the diagonal */
             while (row > 0 && col > 0 && (s1[row-1]==s2[col-1])) {
                             --row;
                             --col;
             }
             blast_d[d][k] = row;

             if (row == 0 && col == 0) {
                 if (d <= max_d) {
                     /* hit northwest corner; have the answer */
                     bput_scr(bscript[d][k],m,n,offset1,offset2,0,0,lblock,rblock);

                     flip_list(lblock,rblock);
                     for (Cost=0; Cost<=MAX_D; Cost++) {
                          free(blast_d[Cost]); free(flast_d[Cost]);
                          free(bscript[Cost]); free(fscript[Cost]);
                     }
                     free(blast_d); free(bscript);
                     free(flast_d); free(fscript);

                     free(min_row); free(min_diag);
                     free(max_row); free(max_diag);
                          ValNodeFreeData(data_list);

                     return d;
                } else goto fin;
             }

             if (row == 0) {
                     /* hit last row; fill in insertions */
                     if (d <= max_d) {
                         new = (struct edit *) ckalloc(sizeof(struct edit));
                         link_to_data_list((Pointer)new, &data_list, &prev);
                         new->line1 = 0;
                         new->line2 = col = --k - ORIGIN;
                         new->op = INSERT;
                         new->link = bscript[d][k+1];
                         bscript[d][k] = new;

                         /* record last INSERT */
                         new = (struct edit *)ckalloc(sizeof(struct edit));
                         link_to_data_list((Pointer)new, &data_list, &prev);
                         new->line1 = 0;
                         new->line2 = col = 0;
                         new->op = INSERT;
                         new->link = bscript[d][k];

                         bput_scr(new,m,n,offset1,offset2,0,0,lblock,rblock);
                         flip_list(lblock,rblock);

                         for (Cost = 0; Cost<=MAX_D; Cost++) {
                              free(blast_d[Cost]); free(flast_d[Cost]);
                              free(bscript[Cost]); free(fscript[Cost]);
                         }
                         free(blast_d); free(bscript);
                         free(flast_d); free(fscript);

                         free(min_row); free(min_diag);
                         free(max_row); free(max_diag);
                              ValNodeFreeData(data_list);

                         return d;

                     } else goto fin;
             }

             if (col == 0) {
                 if (d <= max_d) {
                     /* hit last column; fill in deletions */
                     /* record first DELETE */
                     new = (struct edit *) ckalloc(sizeof(struct edit));
                     link_to_data_list((Pointer)new, &data_list, &prev);
                     new->line1 = -(++k) + ORIGIN;
                     new->line2 = 0;
                     new->op = DELETE;
                     new->link = bscript[d][k-1];
                     bscript[d][k] = new;

                     /* record last DELETE */
                     new = (struct edit *)ckalloc(sizeof(struct edit));
                     link_to_data_list((Pointer)new, &data_list, &prev);
                     new->line1 = 0;
                     new->line2 = 0;
                     new->op = DELETE;
                     new->link = bscript[d][k];

                     bput_scr(new,m,n,offset1,offset2,0,0,lblock,rblock);
                     flip_list(lblock, rblock);

                     for (Cost = 0; Cost<=MAX_D; Cost++) {
                          free(blast_d[Cost]); free(flast_d[Cost]);
                          free(bscript[Cost]); free(fscript[Cost]);
                     }
                     free(blast_d); free(bscript);
                     free(flast_d); free(fscript);

                     free(min_row); free(min_diag);
                     free(max_row); free(max_diag);
                          ValNodeFreeData(data_list);

                     return d;
                 } else goto fin;
             }

        }     /* for k */

        min_row[d] = blast_d[d][DELTA+ORIGIN];
        min_diag[d] = DELTA+ORIGIN;
        for (k=blower; k<=bupper; ++k)
             if (blast_d[d][k]<min_row[d]) {
                       min_row[d] = blast_d[d][k];
                       min_diag[d] = k;
        }

        /* record cell, if paths overlap with minimum combined cost */
        /* obs: it suffices to search up to Cost=min(d-1,(max_d-d)) */
        for (Cost=0; Cost<d; Cost++)
             if ((min_row[d]<=max_row[Cost]) &&
                 (max_d > d+Cost)) {
                           max_d = d+Cost;
                           back = d;
                           forth = Cost;
                           break;
             }

        --blower;
        ++bupper;


        /* for each relevant diagonal ... */
        for (k = flower; k <= fupper; k++) {
               /* get space for the next edit instruction */
               new = (struct edit *) ckalloc(sizeof(struct edit));
                             link_to_data_list((Pointer)new, &data_list, &prev);

               /* find a d on diagonal k */
               if (k==-d+ORIGIN) {
                        /* move down from the last d-1 on diagonal k+1 */
                        row = flast_d[d-1][k+1]+1;
                        new->link = fscript[d-1][k+1];
                        new->op = DELETE;

               } else if (k==d+ORIGIN) {
                        /* move right from the last d-1 on diagonal k-1 */
                        row = flast_d[d-1][k-1];
                        new->link = fscript[d-1][k-1];
                        new->op = INSERT;

               } else if ((flast_d[d-1][k]>=flast_d[d-1][k+1]) &&
                          (flast_d[d-1][k]+1>=flast_d[d-1][k-1])) {

                        /* substitution */
                        row = flast_d[d-1][k]+1;
                        new->link = fscript[d-1][k];
                        new->op = SUBSTITUTE;

               } else if ((flast_d[d-1][k+1]+1>=flast_d[d-1][k-1]) &&
                          (flast_d[d-1][k+1]>=flast_d[d-1][k])) {

                        /* move left from the last d-1 on diagonal k+1 */
                        row = flast_d[d-1][k+1]+1;
                        new->link = fscript[d-1][k+1];
                        new->op = DELETE;
               } else {
                        /* move right from the last d-1 on diagonal k-1 */
                        row = flast_d[d-1][k-1];
                        new->link = fscript[d-1][k-1];
                        new->op = INSERT;

               }
               /* code common to the three cases */
               new->line1 = row;
               new->line2 = col = row + k - ORIGIN;
               fscript[d][k] = new;

               /* slide down the diagonal */
               if (row>=0)
               while (row < m && col < n && (s1[row]==s2[col])) {
                                ++row;
                                ++col;
               }
               flast_d[d][k] = row;

               if (row == m && col == n) {
                   if (d<=max_d) {
                       /* hit southeast corner; have the answer */
                       fput_scr(fscript[d][k],offset1,offset2,m,n,lblock,rblock);

                       for (Cost = 0; Cost<=MAX_D; Cost++) {
                            free(blast_d[Cost]); free(flast_d[Cost]);
                            free(bscript[Cost]); free(fscript[Cost]);
                       }
                       free(blast_d); free(bscript);
                       free(flast_d); free(fscript);

                       free(min_row); free(min_diag);
                       free(max_row); free(max_diag);
                               ValNodeFreeData(data_list);

                       return d;
                   } else goto fin;
               }

               if (row == m) {
                   if (d<=max_d) {
                       /* hit last row; don't look forward */
                       /* record first INSERT */
                       new = (struct edit *)ckalloc(sizeof(struct edit));
                             link_to_data_list((Pointer)new, &data_list, &prev);
                       new->line1 = m;
                       new->line2 = col = m + (++k) - ORIGIN;
                       new->op = INSERT;
                       new->link = fscript[d][k-1];
                       fscript[d][k] = new;

                       /* record last INSERT */
                       new = (struct edit *)ckalloc(sizeof(struct edit));
                             link_to_data_list((Pointer)new, &data_list, &prev);
                       new->line1 = m;
                       new->line2 = n;
                       new->op = INSERT;
                       new->link = fscript[d][k];

                       fput_scr(new,offset1,offset2,m,n,lblock,rblock);

                       for (Cost = 0; Cost<=MAX_D; Cost++) {
                            free(blast_d[Cost]); free(flast_d[Cost]);
                            free(bscript[Cost]); free(fscript[Cost]);
                       }
                       free(blast_d); free(bscript);
                       free(flast_d); free(fscript);

                       free(min_row); free(min_diag);
                       free(max_row); free(max_diag);
                               ValNodeFreeData(data_list);

                       return d;

                   } else goto fin;
               }

               if (col == n) {
                   /* hit last column; don't look forward */
                   if (d<=max_d) {
                       /* record first DELETE */
                       new = (struct edit *)ckalloc(sizeof(struct edit));
                             link_to_data_list((Pointer)new, &data_list, &prev);
                       new->line1 = n - (--k) + ORIGIN;
                       new->line2 = n;
                       new->op = DELETE;
                       new->link = fscript[d][k+1];
                       fscript[d][k] = new;

                       /* record last DELETE */
                       new = (struct edit *)ckalloc(sizeof(struct edit));
                             link_to_data_list((Pointer)new, &data_list, &prev);
                       new->line1 = m;
                       new->line2 = n;
                       new->op = DELETE;
                       new->link = fscript[d][k];

                       fput_scr(new,offset1,offset2,m,n,lblock,rblock);

                       for (Cost = 0; Cost<=MAX_D; Cost++) {
                            free(blast_d[Cost]); free(flast_d[Cost]);
                            free(bscript[Cost]); free(fscript[Cost]);
                       }
                       free(blast_d); free(bscript);
                       free(flast_d); free(fscript);

                       free(min_row); free(min_diag);
                       free(max_row); free(max_diag);
                               ValNodeFreeData(data_list);

                       return d;
                   } else goto fin;
               }
          }     /* for k */


          max_row[d] = flast_d[d][ORIGIN];
          max_diag[d] = ORIGIN;
          for (k=flower; k<=fupper; ++k)
                 if (flast_d[d][k]>max_row[d]) {
                     max_row[d] = flast_d[d][k];
                     max_diag[d] = k;
                 }

          /* record backward and forward limits, if minimum combined
           * cost in overlapping. Note: it suffices to search up to
           * Cost=min(d,(max_d-d)).
           */
          for (Cost=0; Cost<=d; Cost++)
             if ((min_row[Cost]<=max_row[d]) &&
                 (max_d > d+Cost)) {
                      max_d = d+Cost;
                      back = Cost;
                      forth = d;
                      break;
             }
        --flower;
        ++fupper;

        ++d;  /* for d */
      }

  if (d>MAX_D) { 
      *lblock = *rblock = NULL;

      free_this_script (bscript, MAX_D+1);
      free_this_script (fscript, MAX_D+1);
      free_2D_Int(blast_d, MAX_D+1);
      free_2D_Int(flast_d, MAX_D+1);

      free(min_row); free(min_diag);
      free(max_row); free(max_diag);
      ValNodeFreeData(data_list);

      return d;
  }
    
  fin:
        if (m-min_row[back]>=max_row[forth]) {
            bput_scr(bscript[back][min_diag[back]],m,n,
                     offset1,offset2,
                     min_row[back],
                     min_row[back]+min_diag[back]-ORIGIN,
                     &tmp_lblock, &tmp_rblock);
            flip_list(&tmp_lblock,&tmp_rblock);
            *lblock = tmp_lblock;
            *rblock = tmp_rblock;

            fput_scr(fscript[forth][max_diag[forth]],
                     offset1,offset2,
                     min_row[back],
                     min_row[back]+max_diag[forth]-ORIGIN,
                     &tmp_lblock, &tmp_rblock);
            if (*rblock!=NULL) {
                 (*rblock)->next_exon = tmp_lblock;
                 if (tmp_rblock!=NULL) *rblock = tmp_rblock;
            } else {
                 *lblock = tmp_lblock;
                 *rblock = tmp_rblock;
            }
        } else {
            bput_scr(bscript[back][min_diag[back]],m,n,
                     offset1,offset2,
                     max_row[forth],
                     max_row[forth]+min_diag[back]-ORIGIN,
                     &tmp_lblock,&tmp_rblock);
            flip_list(&tmp_lblock,&tmp_rblock);
            *lblock = tmp_lblock;
            *rblock = tmp_rblock;

            fput_scr(fscript[forth][max_diag[forth]],
                     offset1,offset2,
                     max_row[forth],
                     max_row[forth]+max_diag[forth]-ORIGIN,
                     &tmp_lblock, &tmp_rblock);
            if (*rblock!=NULL) {
                 (*rblock)->next_exon = tmp_lblock;
                 if (tmp_rblock!=NULL) *rblock = tmp_rblock;
            } else {
                 *lblock = tmp_lblock;
                 *rblock = tmp_rblock;
            }
        }

        for (Cost=0; Cost<=MAX_D; Cost++) {
                     free(blast_d[Cost]); free(flast_d[Cost]);
                     free(bscript[Cost]); free(fscript[Cost]);
        }
        free(blast_d); free(bscript);
        free(flast_d); free(fscript);
        free(min_row); free(min_diag);
        free(max_row); free(max_diag);
        ValNodeFreeData(data_list);

        return back+forth;

}


static void bput_scr(struct edit *start, int len1, int len2, int offset1, int offset2,int row, int col, Exon **left, Exon **right)
{       struct edit *eptr;
        int ppos1, ppos2;
        Exon *new;

        *right = *left = NULL;

        eptr = start;
        ppos1 = row; ppos2 = col-1;

        while (eptr != NULL) {
           if (ppos1>eptr->line1)        /* not used in this version */
                     eptr = eptr->link;
           else
           switch (eptr->op) {
                case DELETE     :
                case SUBSTITUTE :
                     eptr = eptr->link;
                     break;
                case INSERT     :
                     if (ppos1==eptr->line1) {
                         ppos2 = eptr->line2;
                     } else {
                         new = (Exon *)ckalloc(sizeof(Exon));
                         new->from2 = ppos1+offset1+1;
                         new->to2 = eptr->line1-1+offset1+1;
                         new->from1 = ppos2+1+offset2+1;
                         new->to1 = eptr->line2-1+offset2+1;
                         new->flag = 0;
                         new->length = new->to2-new->from2+1;
                         new->next_exon = NULL;
                         if (*right==NULL)
                             *right = *left = new;
                         else {
                            (*right)->next_exon = new;
                             *right = new;
                         }

                         ppos1 = eptr->line1;
                         ppos2 = eptr->line2;
                     }
                     eptr=eptr->link;
           }
        }
        if ((ppos1!=len1) || (ppos2+1!=len2)) {

                  new = (Exon *)ckalloc(sizeof(Exon));
                  new->from2 = ppos1+offset1+1;
                  new->to2 = len1-1+offset1+1;
                  new->from1 = ppos2+1+offset2+1;
                  new->to1 = len2-1+offset2+1;
                  new->flag = 0;
                  new->length = new->to2-new->from2+1;
                  new->next_exon = NULL;
                  if (*right==NULL)
                      *right = *left = new;
                  else {
                     (*right)->next_exon = new;
                      *right = new;
                  }

        }
}

/* reverse global ... */
static void fput_scr(struct edit *start, int offset1, int offset2, int row, int col, Exon **left, Exon **right)
{       struct edit *eptr;
        int    ppos1, ppos2;
        Exon  *new;

        *left = *right = NULL;

        eptr = start;
        ppos1 = row; ppos2 = col+1;

        while (eptr != NULL) {
           if (ppos1 < eptr->line1)        /* not used in this version */
                eptr = eptr->link;
           else
           switch (eptr->op) {
                case DELETE     :
                case SUBSTITUTE :
                     eptr = eptr->link;
                     break;
                case INSERT     :
                     if (ppos1==eptr->line1) {
                         ppos2 = eptr->line2;
                     } else {
                         new = (Exon *)ckalloc(sizeof(Exon));
                         new->from2 = eptr->line1+1+offset1;
                         new->to2 = ppos1+offset1;
                         new->from1 = eptr->line2+1+offset2;
                         new->to1 = ppos2-1+offset2;
                         new->flag = 0;
                         new->length = new->to2-new->from2+1;
                         new->next_exon = NULL;
                         if (*right==NULL)
                             *right = *left = new;
                         else { 
                            (*right)->next_exon = new;
                             *right = new;
                         }

                         ppos1 = eptr->line1;
                         ppos2 = eptr->line2;
                     }
                     eptr=eptr->link;
           }
        }
        if (ppos1!=0) {  /* not a final insert */
                  new = (Exon *)ckalloc(sizeof(Exon));
                  new->from2 = 1+offset1;
                  new->to2 = ppos1+offset1;
                  new->from1 = 1+offset2;
                  new->to1 = ppos2-1+offset2;
                  new->flag = 0;
                  new->length = new->to2-new->from2+1;
                  new->next_exon = NULL;
                  if (*right==NULL)
                      *right = *left = new;
                  else {
                     (*right)->next_exon = new;
                      *right = new;
                  }
        }
}

static void flip_list(Exon **left, Exon **right)
{
   Exon   *ep, *ahead, *behind;

   *right = *left;
   ahead = *left;
   ep = NULL;
   while (ahead!=NULL) {
          behind = ep;
          ep = ahead;
          ahead = ahead->next_exon;
          ep->next_exon = behind;
  }
  *left = ep;
}


/* operates on a list sorted in increasing order of exon coordinates */
static void compact_list(Exon **Lblock, Exon **Rblock)
{  
     Exon *tmp_block=*Lblock, *tmp_block1;

     while ((tmp_block!=NULL) && 
            ((tmp_block1=tmp_block->next_exon)!=NULL) &&
            tmp_block1->to1) {
        if ((abs((tmp_block1->from2-tmp_block1->from1) -
                (tmp_block->to2-tmp_block->to1))<=W) &&
                (tmp_block1->from2-tmp_block->to2-1<=MAX_INTERNAL_GAP)) {
            /* merge blocks */
            tmp_block->to1 = tmp_block1->to1;
            tmp_block->to2 = tmp_block1->to2;
            tmp_block->length = tmp_block->to2-tmp_block->from2+1;
            tmp_block->next_exon = tmp_block1->next_exon;
    
            free(tmp_block1);
        } else
            tmp_block = tmp_block1;
     }
     /* reset right end of the list */
     *Rblock = tmp_block;
}

/* ------------------  memory management routines --------------- */

void link_to_data_list(Pointer data, ValNodePtr *head, ValNodePtr *prev)
{
     ValNodePtr curr;

     curr = (ValNodePtr)ckalloc(sizeof(struct ValNode));
     curr->data = data;
     curr->next = NULL;

     if(*prev == NULL)  
        *head = curr;
     else
        (*prev)->next = curr;
     *prev = curr;
}

void   ValNodeFreeData(ValNodePtr data_list)
{
       ValNodePtr   tmp_node;

       while ((tmp_node=data_list)!=NULL) {
          free(tmp_node->data);
          data_list = data_list->next;
          free(tmp_node); 
       }
}
  
      

static void free_2D_Int(int ** val, int size)
{
        int i;

        for(i=0; i<size; ++i)
            free(val[i]);
        free(val);
}


static void free_this_script (struct edit ***val, int size)
{
        int i;

        for(i=0; i<size; ++i)
            free(val[i]);
        free(val);
}

int  good_ratio(int length)
{
     if (length<=W/2) return 2;
     else if (length<2*W) return cutoff;
     else return (int)(P*length+1);
}

static int extend_bw(uchar *s1, uchar *s2, int m, int n, int offset1, int offset2, int *line1, int *line2)
{
  int     col,                    /* column number */
          row,                    /* row number */
          max_d,                  /* bound on the length of the edit script
 */
          d,                      /* current compressed distance */
          k,                      /* current diagonal */
          DELTA,                  /* n-m  */
          ORIGIN,
          lower,
          upper;
  int     *last_d, *temp_d;       /* column containing the last p */
  int     *min_row, *min_diag;    /* min (b)/ max (f) row (and diagonal) */
                                  /* reached for cost d=0, ... m.  */
  DELTA = n-m;
  max_d = m+1;

  ORIGIN = m;
  for (row=m, col=n; row>0 && col>0 && (s1[row-1]==s2[col-1]); row--,col--)
        /*LINTED empty loop body*/; 

  if ((row == 0) || (col == 0)) {
       *line1 = row+offset1;
       *line2 = col+offset2;

       return 0;
  }

  last_d = (int *)ckalloc((m+n+1)*sizeof(int));
  temp_d = (int *)ckalloc((m+n+1)*sizeof(int));

  for (k=0; k<=m+n; ++k) last_d[k]=m+1;
  last_d[ORIGIN+DELTA] = row;

  lower = ORIGIN + DELTA - 1;
  upper = ORIGIN + DELTA + 1;

  min_row = (int *)ckalloc((m+1)*sizeof(int));
  min_diag = (int *)ckalloc((m+1)*sizeof(int));

  for (d=1; d<=m; d++)
       min_row[d] = m+1;
 
  min_row[0] = last_d[ORIGIN+DELTA];
  min_diag[0] = ORIGIN + DELTA;

  d = 0;
  while ((++d<=max_d) && 
         ((d-1<=good_ratio(m-min_row[d-1])) ||
          ((d>=2) && (d-2<=good_ratio(m-min_row[d-2]))))) {

          /* for each relevant diagonal ... */
          for (k = lower; k <= upper; k++) {

               /* find a d on diagonal k */
               if (k==-d+DELTA+ORIGIN) {
                        /* move down from the last d-1 on diagonal k+1 */
                        row = last_d[k+1];
                        /* op = INSERT; */

               } else if (k==d+DELTA+ORIGIN) {
                        /* move right from the last d-1 on diagonal k-1 */
                        row = last_d[k-1]-1;
                        /* op = DELETE; */

               } else if ((last_d[k]-1<=last_d[k+1]) &&
                          (last_d[k]-1<=last_d[k-1]-1)) {
                        /* substitution */
                        row = last_d[k]-1;
                        /* op = SUBSTITUTE; */

               } else if ((last_d[k-1]-1<=last_d[k+1]) &&
                          (last_d[k-1]-1<=last_d[k]-1)) {
                        /* move right from the last d-1 on diagonal k-1 */
                        row = last_d[k-1]-1;
                        /* op = DELETE; */

               } else  {
                        /* move left from the last d-1 on diagonal k+1 */
                        row = last_d[k+1];
                        /* op = INSERT; */

               }

               /* code common to the three cases */
               /* slide down the diagonal */

               col = row+k-ORIGIN;

               while ((row > 0) && (col > 0) && (s1[row-1]==s2[col-1]))
                  { row--; col--; }

               temp_d[k] = row;

               if ((row == 0) && (col == 0)) {
                   /* hit southeast corner; have the answer */

                   free(last_d); free(temp_d);
                   free(min_row); free(min_diag);

                   *line1 = row+offset1;
                   *line2 = col+offset2;

                   return d;
               }
               if (row == 0) {
                   /* hit first row; don't look further */

                   free(last_d); free(temp_d);
                   free(min_row); free(min_diag);

                   *line1 = row+offset1;
                   *line2 = col+offset2;

                   return d;
               }

               if (col == 0) {
                   /* hit last column; don't look further */
                   free(last_d); free(temp_d);
                   free(min_row); free(min_diag);

                   *line1 = row+offset1;
                   *line2 = col+offset2;

                   return d;
               }
          }

          min_row[d] = last_d[ORIGIN+DELTA];
          min_diag[d] = ORIGIN+DELTA;
          for (k=lower; k<=upper; ++k)
                 if (temp_d[k]<min_row[d]) {
                     min_row[d] = temp_d[k];
                     min_diag[d] = k;
                 }

          for (k=lower; k<=upper; k++) {
               last_d[k] = temp_d[k];
          }

          --lower;
          ++upper;
     }

    /* report here the previous maximal match, stored in min_diag and min_row */
     while ((d>0) && (min_row[d-1]-min_row[d]<3))
        d--;

     *line1 = min_row[d]+offset1;
     *line2 = min_row[d]+min_diag[d]-ORIGIN+offset2;

     free(min_row);
     free(min_diag);
     free(last_d);
     free(temp_d);

     return d;
}


static int extend_fw(uchar *s1, uchar *s2, int m, int n, int offset1, int offset2, int *line1, int *line2)
{
  int     col,                    /* column number */
          row,                    /* row number */
          max_d,                  /* bound on the length of the edit script
 */
          d,                      /* current compressed distance */
          k,                      /* current diagonal */
          ORIGIN,
          lower,
          upper;
  int     *last_d, *temp_d;       /* column containing the last p */
  int     *max_row, *max_diag;    /* min (b)/ max (f) row (and diagonal) */
                                  /* reached for cost d=0, ... m.  */
  max_d = m+1;

  ORIGIN = m;
  for (row=0, col=0; col<n && row<m && (s1[row]==s2[col]); row++, col++)
        /*LINTED empty loop body*/; 

  if (row == m) {
       *line1 = row+offset1;
       *line2 = col+offset2;

       return 0;
  }
  if (col == n) {
       *line1 = row+offset1;
       *line2 = col+offset2;

       return 0;
  }

  last_d = (int *)ckalloc((m+n+1)*sizeof(int));
  temp_d = (int *)ckalloc((m+n+1)*sizeof(int));

  for (k=0; k<=m+n; ++k) last_d[k]=-1;
  last_d[ORIGIN] = row;

  lower = ORIGIN - 1;
  upper = ORIGIN + 1;

  max_row = (int *)ckalloc((m+1)*sizeof(int));
  max_diag = (int *)ckalloc((m+1)*sizeof(int));

  for (d=1; d<=m; d++)
       max_row[d] = -1;
 
  max_row[0] = last_d[ORIGIN];
  max_diag[0] = ORIGIN;


  d = 0;
  while ((++d<=max_d) && 
         ((d-1<=good_ratio(max_row[d-1])) || 
          ((d>=2) && (d-2<=good_ratio(max_row[d-2]))))) {

          /* for each relevant diagonal ... */
          for (k = lower; k <= upper; k++) {

               /* find a d on diagonal k */
               if (k==-d+ORIGIN) {

                        /* move down from the last d-1 on diagonal k+1 */
                        row = last_d[k+1]+1;
                        /* op = DELETE; */
               } else if (k==d+ORIGIN) {

                        /* move right from the last d-1 on diagonal k-1 */
                        row = last_d[k-1];
                        /* op = INSERT; */
               } else if ((last_d[k]>=last_d[k+1]) &&
                          (last_d[k]+1>=last_d[k-1])) {

                        /* substitution */
                        row = last_d[k]+1;
                        /* op = SUBSTITUTE; */
               } else if ((last_d[k+1]+1>=last_d[k-1]) &&
                          (last_d[k+1]>=last_d[k])) {

                        /* move down from the last d-1 on diagonal k+1 */
                        row = last_d[k+1]+1;
                        /* op = DELETE; */
               } else {

                        /* move right from the last d-1 on diagonal k-1 */
                        row = last_d[k-1];
                        /* op = INSERT; */
               }

               /* code common to the three cases */
               /* slide down the diagonal */

               col = row+k-ORIGIN;

               if (row>=0)
               while ((row < m) && (col < n) && (s1[row]==s2[col]))
                         { row++; col++; }

               temp_d[k] = row;

               if ((row == m) && (col == n)) {
                   /* hit southeast corner; have the answer */
                   free(last_d); free(temp_d);
                   free(max_row); free(max_diag);
                   *line1 = row+offset1;
                   *line2 = col+offset2;

                   return d;
               }
               if (row == m) {
                   /* hit last row; don't look further */
                   free(temp_d); free(last_d);
                   free(max_row); free(max_diag);

                   *line1 = row+offset1;
                   *line2 = col+offset2;

                   return d;
               }

               if (col == n) {
                   /* hit last column; don't look further */
                   free(temp_d); free(last_d);
                   free(max_row); free(max_diag);

                   *line1 = row+offset1;
                   *line2 = col+offset2;

                   return d;
               }
          }
          max_row[d] = last_d[ORIGIN];
          max_diag[d] = ORIGIN;
          for (k=lower; k<=upper; ++k)
                 if (temp_d[k]>max_row[d]) {
                     max_row[d] = temp_d[k];
                     max_diag[d] = k;
                 }

          for (k=lower; k<=upper; k++) {
               last_d[k] = temp_d[k];
          }

          --lower;
          ++upper;
     }

    /* report here the previous maximal match, stored in max_diag and max_row */

     while ((d>0) && (max_row[d]-max_row[d-1]<3))
        d--;

     *line1 = max_row[d]+offset1;
     *line2 = max_row[d]+max_diag[d]-ORIGIN+offset2;

     free(max_row);
     free(max_diag);
     free(last_d);
     free(temp_d);

     return d;

/*
     if ((d>2) && (max_row[d-1]-max_row[d-2]<3)) {
          *line1 = max_row[d-2]+offset1;
          *line2 = max_row[d-2]+max_diag[d-2]-ORIGIN+offset2;
 
          free(max_row); free(max_diag);
          free(last_d); free(temp_d);
 
          return d-2;
     }
 
     *line1 = max_row[d-1]+offset1;
     *line2 = max_row[d-1]+max_diag[d-1]-ORIGIN+offset2;

     free(max_row);
     free(max_diag);
     free(last_d);
     free(temp_d);

     return d-1;
*/
}

static  void merge(Exon **t0, Exon **t1) 
{   
    Exon  *tmp0, *tmp1;

    if ((*t0) && !(*t0)->to1)
         tmp0 = (*t0)->next_exon;
    else    
         tmp0 = *t0;

    while (tmp0 && (tmp0!=*t1)) {
       tmp1 = tmp0->next_exon;
       if ((abs((tmp1->from2-tmp1->from1)-
                (tmp0->to2-tmp0->to1))<=W) &&
           (tmp1->from2-tmp0->to2<=W)) {
         
          /* merge blocks tmp0 and tmp1 */
          tmp0->to1 = tmp1->to1;
          tmp0->to2 = tmp1->to2;
          tmp0->length = tmp0->to2-tmp0->from2+1;
          if (tmp1==*t1) {
              tmp0->flag = (*t1)->flag;
              *t1 = tmp0;
          }
          tmp0->next_exon = tmp1->next_exon;  

          free(tmp1);
       } else 
          tmp0 = tmp0->next_exon;
    }
}


/*
static void slide(int w, char *ori1, int *xt0, int *xt1, int *yt0, int *yt1)
{      
       uchar *s, *t;
       int    i, C_score, G_score, best_G=0, best_C=0, C_off, G_off;
            
       if (*(seq1+*xt0)=='G')       best_G++;
       else 
           if (*(seq1+*xt0)=='C')   best_C++;
       if (*(seq1+*xt0+1)=='T')   { best_G++; best_C++;}
       if (*(seq1+*xt1-3)=='A')   { best_G++; best_C++;}
       if (*(seq1+*xt1-2)=='G')     best_G++;
       else
           if (*(seq1+*xt1-2)=='C') best_C++;
           
       C_off = G_off = 0;
 
       s = seq1+*xt0-w-1;
       t = seq1+*xt1-w-1;
       for (i=-w; i<=w; i++,++s,++t) {
            G_score = C_score = 0;
            if (*(s+1)=='G')  G_score++;
            else if (*(s+1)=='C') C_score++;
            if (*(s+2)=='T') { G_score++; C_score++; }
            if (*(t-2)=='A') { G_score++; C_score++; }
            if (*(t-1)=='G')  G_score++;
            else if (*(t-1)=='C') C_score++;
            if (C_score>best_C) { 
                best_C = C_score; 
                C_off = i;
            }   
            if (G_score>best_G) {
                best_G = G_score;
                G_off = i;
            }   
       }    
            
       if (best_C > best_G) {
            *xt1 += C_off;
            *xt0 += C_off;
            *yt1 += C_off;
            *yt0 += C_off;

            if (best_C>=3) *ori1 = 'C';
            else *ori1 = 'N';
       } else {
            *xt1 += G_off;
            *xt0 += G_off;
            *yt1 += G_off;
            *yt0 += G_off;

            if (best_G>=3) *ori1 = 'G';
            else *ori1 = 'N';
       }    
}
*/

static void slide(int w, char *ori1, int *xt0, int *xt1, int *yt0, int *yt1, int *score)
{
       uchar *s, *t, *s0, *t0;
       int    best_x=0, x, i, j, C_score, G_score, best_G=0, best_C=0, C_off, G_off;

       s = seq1+*xt0-1; t = seq1+*xt1-1;
       s0 = seq2+*yt0-1; t0 = seq2+*yt1-1;
       for (i=0; i<w; i++) {
            best_x += ((*s0--!=*s--) ? 1 : 0);
            best_x += ((*t0++!=*t++) ? 1 : 0);
       }

       if (*(seq1+*xt0)=='G')       best_G++;
       else
           if (*(seq1+*xt0)=='C')   best_C++;
       if (*(seq1+*xt0+1)=='T')   { best_G++; best_C++;}
       if (*(seq1+*xt1-3)=='A')   { best_G++; best_C++;}
       if (*(seq1+*xt1-2)=='G')     best_G++; 
       else
           if (*(seq1+*xt1-2)=='C') best_C++;
           
       C_off = G_off = 0;
 
       s = seq1+*xt0-w-1;  s0 = seq2+*yt0-w-1;
       t = seq1+*xt1-w-1;  t0 = seq2+*yt1-w-1;
 
       x = -1;
       for (i=-w; i<=w; i++,++s,++t,++s0,++t0) {
            if (x<0) {
                for (j=x=0; j<2*w; j++)
                    x += ((*(t0+j)==*(t+j)) ? 0:1);
            } else {
                x -= ((*(t0-1)==*(t-1)) ? 0:1);
                x += ((*s0==*s) ? 0:1);
            }   
            G_score = C_score = 0;
            if (*(s+1)=='G')  G_score++;
            else if (*(s+1)=='C') C_score++;
            if (*(s+2)=='T') { G_score++; C_score++; }
            if (*(t-2)=='A') { G_score++; C_score++; }
            if (*(t-1)=='G')  G_score++;
            else if (*(t-1)=='C') C_score++;
            if ((C_score>best_C) && (best_x-x>=-1)) {
                best_C = C_score;
                if (best_x<x) best_x = x;
                C_off = i;
            }   
            if ((G_score>best_G) && (best_x-x>=-1)) {
                best_G = G_score;
                if (best_x<x) best_x = x;
                G_off = i;
            }   
       }    
            
       if (best_C > best_G) {
            *xt1 += C_off;
            *xt0 += C_off;
            *yt1 += C_off;
            *yt0 += C_off;
            
            if (best_C>=3) *ori1 = 'C';
            else *ori1 = 'N';
            *score = -best_C;
       } else {
            *xt1 += G_off;
            *xt0 += G_off;
            *yt1 += G_off;
            *yt0 += G_off;
            
            if (best_G>=3) *ori1 = 'G';
            else *ori1 = 'N';
            *score = best_G;
       }    
}
 

void free_align(edit_script_list *aligns)
{
     edit_script_list *head;

     head = aligns;

     while ((head=aligns)!=NULL) {
        aligns = aligns->next_script;
        Script_free(head->script);
        free(head); 
     }
}

     
void free_list(Exon *left)
{     
     Exon *tmp_block;
      
     while ((tmp_block=left)!=NULL) {
          left = left->next_exon;
          free(tmp_block);
     }
}    

void free_table(void)
{    
     register struct hash_node *hptr, *tptr;
     register int    hval;

     free(pnext_pos);
     for (hval=0; hval<HASH_SIZE+1; hval++) {
          hptr = phashtab[hval];
          while (hptr) {
             tptr = hptr;
             hptr = hptr->link;
             free(tptr);
          } 
     }     
}  

static Exon *bmatch (uchar *s1, uchar *s2, int len1, int len2, int offset1, int offset2)
{
       int  i, j, i1, score;
       Exon *new=NULL;
 
       for (i1=i=len1-3; i>=len2-3; i--, i1=i) {
         for (j=len2-3; j>=2; j--, i1--)
           if (*(s1+i1)!=*(s2+j))
               break;
          
         if (j<2) {
            /* exact match for CDS found; check signals */
            score = 0;
            if (*(s1+(i1--))==*(s2+(j--))) score++;
            if (*(s1+(i1--))==*(s2+(j--))) score++;
            if (*(s1+i1+len2-1)==*(s2+j+len2-1)) score++;
            if (*(s1+i1+len2)==*(s2+j+len2)) score++;
            if (score>=3) {
                new = (Exon *)ckalloc(sizeof(Exon));
                new->from1 = i1+3+offset1;
                new->from2 = offset2;
                new->to1 = new->from1+len2-5;
                new->to2 = new->from2+len2-5;
                new->ori = (G_score >= abs(C_score)) ? 'G' : 'C';
                new->length = len2-4; 
                new->next_exon = NULL;
    
                return new;
            }
         }
       }
       return NULL;
}

static Exon *fmatch (uchar *s1, uchar *s2, int len1, int len2, int offset1, int offset2)
{
       int  i, j, i1, score;
       Exon *new=NULL;
 
       for (i1=i=2; i<len1-len2+3; i++, i1=i) {
         for (j=2; j<len2-2; j++, i1++)
           if (*(s1+i1)!=*(s2+j))
               break;

         if (j>=len2-2) {
            /* exact match found for internal part, look for signals */
            score = 0; 
            if (*(s1+(i1++))==*(s2+(j++))) score++;
            if (*(s1+(i1++))==*(s2+(j++))) score++;
            if (*(s1+i1-len2)==*s2) score++;
            if (*(s1+i1-len2+1)==*(s2+1)) score++;
            if (score>=3) {
                new = (Exon *)ckalloc(sizeof(Exon));
                new->from1 = i+offset1;
                new->to1 = i1+offset1-2;
                new->from2 = offset2;
                new->to2 = new->from2+len2-5;
                new->ori = (G_score >= abs(C_score)) ? 'G' : 'C';
                new->length = len2-4;
                new->next_exon = NULL;

                return new;
            }
         }  
       }    
       return NULL;
}      
